"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0060.pubsub import XEP_0060
from slixmpp.plugins.xep_0060 import stanza


register_plugin(XEP_0060)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""
from typing import (
    Dict,
    Optional,
    Set,
    Tuple,
)

from slixmpp import JID, Message, Iq
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.matcher import MatchXPath
from slixmpp.xmlstream.handler import Callback
from slixmpp.plugins.xep_0404 import stanza
from slixmpp.plugins.xep_0004.stanza import Form


NODES = [
    'urn:xmpp:mix:nodes:jidmap',
]


class XEP_0404(BasePlugin):
    '''XEP-0404: MIX JID Hidden Channels'''

    name = 'xep_0404'
    description = 'MIX-ANON'
    dependencies = {'xep_0369'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()

    async def get_anon_raw(self, channel: JID, *,
                            ifrom: Optional[JID] = None, **pubsubkwargs) -> Iq:
        """
        Get the jid-participant mapping result (raw).
        :param JID channel: MIX channel JID
        """
        return await self.xmpp['xep_0030'].get_items(
            channel.bare,
            ifrom=ifrom,
            **pubsubkwargs
        )

    async def get_anon_by_jid(self, channel: JID, *,
                              ifrom: Optional[JID] = None, **pubsubkwargs) -> Dict[JID, str]:
        """
        Get the jid-participant mapping, by JID

        :param JID channel: MIX channel JID
        """
        raw = await self.get_anon_raw(channel, ifrom=ifrom, **pubsubkwargs)
        mapping = {}
        for item in raw['pubsub']['items']:
            mapping[item['anon_participant']['jid']] = item['id']
        return mapping

    async def get_anon_by_id(self, channel: JID, *,
                             ifrom: Optional[JID] = None, **pubsubkwargs) -> Dict[str, JID]:
        """
        Get the jid-participant mapping, by participant id

        :param JID channel: MIX channel JID
        """
        raw = await self.get_anon_raw(channel, ifrom=ifrom, **pubsubkwargs)
        mapping = {}
        for item in raw['pubsub']['items']:
            mapping[item['id']] = item['anon_participant']['jid']
        return mapping

    async def get_preferences(self, channel: JID, *,
                              ifrom: Optional[JID] = None, **iqkwargs) -> Form:
        """
        Get channel preferences with default values.
        :param JID channel: MIX channel JID
        """
        iq = self.xmpp.make_iq_get(ito=channel.bare, ifrom=ifrom)
        iq.enable('user_preference')
        prefs_stanza = await iq.send(**iqkwargs)
        return prefs_stanza['user_preference']['form']

    async def set_preferences(self, channel: JID, form: Form, *,
                              ifrom: Optional[JID] = None, **iqkwargs) -> Form:
        """
        Set channel preferences
        :param JID channel: MIX channel JID
        :param Form form: A 0004 form with updated preferences
        """
        iq = self.xmpp.make_iq_set(ito=channel.bare, ifrom=ifrom)
        iq['user_preference']['form'] = form
        prefs_result = await iq.send(**iqkwargs)
        return prefs_result['user_preference']['form']



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0352.stanza import Active, Inactive, ClientStateIndication
from slixmpp.plugins.xep_0352.csi import XEP_0352


register_plugin(XEP_0352)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0012.stanza import LastActivity
from slixmpp.plugins.xep_0012.last_activity import XEP_0012


register_plugin(XEP_0012)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.stanza import Iq
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.google.settings import stanza


class GoogleSettings(BasePlugin):

    """
    Google: Gmail Notifications

    Also see <https://developers.google.com/talk/jep_extensions/usersettings>.
    """

    name = 'google_settings'
    description = 'Google: User Settings'
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, stanza.UserSettings)

        self.xmpp.register_handler(
                Callback('Google Settings',
                    StanzaPath('iq@type=set/google_settings'),
                    self._handle_settings_change))

    def plugin_end(self):
        self.xmpp.remove_handler('Google Settings')

    def get(self, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq.enable('google_settings')
        return iq.send(timeout=timeout, callback=callback)

    def update(self, settings, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq.enable('google_settings')

        for setting, value in settings.items():
            iq['google_settings'][setting] = value

        return iq.send(timeout=timeout, callback=callback)

    def _handle_settings_change(self, iq):
        reply = self.xmpp.Iq()
        reply['type'] = 'result'
        reply['id'] = iq['id']
        reply['to'] = iq['from']
        reply.send()
        self.xmpp.event('google_settings_change', iq)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2016 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0334 import stanza, Store, NoStore, NoPermanentStore, NoCopy

log = logging.getLogger(__name__)

class XEP_0334(BasePlugin):

    name = 'xep_0334'
    description = 'XEP-0334: Message Processing Hints'
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, Store)
        register_stanza_plugin(Message, NoStore)
        register_stanza_plugin(Message, NoPermanentStore)
        register_stanza_plugin(Message, NoCopy)



# This module is a very stripped down version of the dateutil
# package for when dateutil has not been installed. As a replacement
# for dateutil.parser.parse, the parsing methods from
# http://blog.mfabrik.com/2008/06/30/relativity-of-time-shortcomings-in-python-datetime-and-workaround/

#As such, the following copyrights and licenses applies:


# dateutil - Extensions to the standard python 2.3+ datetime module.
#
# Copyright (c) 2003-2011 - Gustavo Niemeyer <gustavo@niemeyer.net>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# fixed_dateime
#
# Copyright (c) 2008, Red Innovation Ltd., Finland
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Red Innovation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED INNOVATION ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL RED INNOVATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import re
import math
import datetime


ZERO = datetime.timedelta(0)


try:
    from dateutil.parser import parse as parse_iso
    from dateutil.tz import tzoffset, tzutc
except:
    # As a stopgap, define the two timezones here based
    # on the dateutil code.

    class tzutc(datetime.tzinfo):

        def utcoffset(self, dt):
            return ZERO

        def dst(self, dt):
            return ZERO

        def tzname(self, dt):
            return "UTC"

        def __eq__(self, other):
            return (isinstance(other, tzutc) or
                    (isinstance(other, tzoffset) and other._offset == ZERO))

        def __ne__(self, other):
            return not self.__eq__(other)

        def __repr__(self):
            return "%s()" % self.__class__.__name__

        __reduce__ = object.__reduce__

    class tzoffset(datetime.tzinfo):

        def __init__(self, name, offset):
            self._name = name
            self._offset = datetime.timedelta(minutes=offset)

        def utcoffset(self, dt):
            return self._offset

        def dst(self, dt):
            return ZERO

        def tzname(self, dt):
            return self._name

        def __eq__(self, other):
            return (isinstance(other, tzoffset) and
                    self._offset == other._offset)

        def __ne__(self, other):
            return not self.__eq__(other)

        def __repr__(self):
            return "%s(%s, %s)" % (self.__class__.__name__,
                                   repr(self._name),
                                   self._offset.days*86400+self._offset.seconds)

        __reduce__ = object.__reduce__


    _fixed_offset_tzs = { }
    UTC = tzutc()

    def _get_fixed_offset_tz(offsetmins):
        """For internal use only: Returns a tzinfo with
        the given fixed offset. This creates only one instance
        for each offset; the zones are kept in a dictionary"""

        if offsetmins == 0:
            return UTC

        if not offsetmins in _fixed_offset_tzs:
            if offsetmins < 0:
                sign = '-'
                absoff = -offsetmins
            else:
                sign = '+'
                absoff = offsetmins

            name = "UTC%s%02d:%02d" % (sign, int(absoff / 60), absoff % 60)
            inst = tzoffset(name,offsetmins)
            _fixed_offset_tzs[offsetmins] = inst

        return _fixed_offset_tzs[offsetmins]


    _iso8601_parser = re.compile(r"""
        ^
        (?P<year> [0-9]{4})?(?P<ymdsep>-?)?
        (?P<month>[0-9]{2})?(?P=ymdsep)?
        (?P<day>  [0-9]{2})?

        (?P<time>
            (?: # time part... optional... at least hour must be specified
            (?:T|\s+)?
                (?P<hour>[0-9]{2})
                (?:
                    # minutes, separated with :, or none, from hours
                    (?P<hmssep>[:]?)
                    (?P<minute>[0-9]{2})
                    (?:
                        # same for seconds, separated with :, or none, from hours
                        (?P=hmssep)
                        (?P<second>[0-9]{2})
                    )?
                )?

                # fractions
                (?: [,.] (?P<frac>[0-9]{1,10}))?

                # timezone, Z, +-hh or +-hh:?mm. MUST BE, but complain if not there.
                (
                    (?P<tzempty>Z)
                |
                    (?P<tzh>[+-][0-9]{2})
                    (?: :? # optional separator
                        (?P<tzm>[0-9]{2})
                    )?
                )?
            )
        )?
        $
    """, re.X) # """

    def parse_iso(timestamp):
        """Internal function for parsing a timestamp in
        ISO 8601 format"""

        timestamp = timestamp.strip()

        m = _iso8601_parser.match(timestamp)
        if not m:
            raise ValueError("Not a proper ISO 8601 timestamp!: %s" % timestamp)

        vals = m.groupdict()
        def_vals = {'year': 1970, 'month': 1, 'day': 1}
        for key in vals:
            if vals[key] is None:
                vals[key] = def_vals.get(key, 0)
            elif key not in ['time', 'ymdsep', 'hmssep', 'tzempty']:
                vals[key] = int(vals[key])

        year  = vals['year']
        month = vals['month']
        day   = vals['day']

        if m.group('time') is None:
            return datetime.date(year, month, day)

        h, min, s, us = None, None, None, 0
        frac = 0
        if m.group('tzempty') == None and m.group('tzh') == None:
            raise ValueError("Not a proper ISO 8601 timestamp: " +
                    "missing timezone (Z or +hh[:mm])!")

        if m.group('frac'):
            frac = m.group('frac')
            power = len(frac)
            frac  = int(frac) / 10.0 ** power

        if m.group('hour'):
            h = vals['hour']

        if m.group('minute'):
            min = vals['minute']

        if m.group('second'):
            s = vals['second']

        if frac != None:
            # ok, fractions of hour?
            if min == None:
                frac, min = math.modf(frac * 60.0)
                min = int(min)

            # fractions of second?
            if s == None:
                frac, s = math.modf(frac * 60.0)
                s = int(s)

            # and extract microseconds...
            us = int(frac * 1000000)

        if m.group('tzempty') == 'Z':
            offsetmins = 0
        else:
            # timezone: hour diff with sign
            offsetmins = vals['tzh'] * 60
            tzm = m.group('tzm')

            # add optional minutes
            if tzm != None:
                tzm = int(tzm)
                offsetmins += tzm if offsetmins > 0 else -tzm

        tz = _get_fixed_offset_tz(offsetmins)
        return datetime.datetime(year, month, day, h, min, s, us, tz)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010  Nathanael C. Fritz
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.xmlstream.matcher.id import MatcherId
from slixmpp.xmlstream.matcher.idsender import MatchIDSender
from slixmpp.xmlstream.matcher.many import MatchMany
from slixmpp.xmlstream.matcher.stanzapath import StanzaPath
from slixmpp.xmlstream.matcher.xmlmask import MatchXMLMask
from slixmpp.xmlstream.matcher.xpath import MatchXPath

__all__ = ['MatcherId', 'MatchMany', 'StanzaPath',
           'MatchXMLMask', 'MatchXPath']



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.google.auth import stanza


class GoogleAuth(BasePlugin):

    """
    Google: Auth Extensions (JID Domain Discovery, OAuth2)

    Also see:
        <https://developers.google.com/talk/jep_extensions/jid_domain_change>
        <https://developers.google.com/talk/jep_extensions/oauth>
    """

    name = 'google_auth'
    description = 'Google: Auth Extensions (JID Domain Discovery, OAuth2)'
    dependencies = set(['feature_mechanisms'])
    stanza = stanza

    def plugin_init(self):
        self.xmpp.namespace_map['http://www.google.com/talk/protocol/auth'] = 'ga'

        register_stanza_plugin(self.xmpp['feature_mechanisms'].stanza.Auth,
                               stanza.GoogleAuth)

        self.xmpp.add_filter('out', self._auth)

    def plugin_end(self):
        self.xmpp.del_filter('out', self._auth)

    def _auth(self, stanza):
        if isinstance(stanza, self.xmpp['feature_mechanisms'].stanza.Auth):
            stanza.stream = self.xmpp
            stanza['google']['client_uses_full_bind_result'] = True
            if stanza['mechanism'] == 'X-OAUTH2':
                stanza['google']['service'] = 'oauth2'
            print(stanza)
        return stanza



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0009 import stanza
from slixmpp.plugins.xep_0009.rpc import XEP_0009
from slixmpp.plugins.xep_0009.stanza import RPCQuery, MethodCall, MethodResponse


register_plugin(XEP_0009)



import asyncio
import logging
import socket

from hashlib import sha1
from uuid import uuid4

from slixmpp.stanza import Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins.base import BasePlugin

from slixmpp.plugins.xep_0065 import stanza, Socks5, Socks5Protocol


log = logging.getLogger(__name__)


class XEP_0065(BasePlugin):

    name = 'xep_0065'
    description = "XEP-0065: SOCKS5 Bytestreams"
    dependencies = {'xep_0030'}
    default_config = {
        'auto_accept': False
    }

    def plugin_init(self):
        register_stanza_plugin(Iq, Socks5)

        self._proxies = {}
        self._sessions = {}
        self._preauthed_sids = {}

        self.xmpp.register_handler(
            Callback('Socks5 Bytestreams',
                     StanzaPath('iq@type=set/socks/streamhost'),
                     self._handle_streamhost))

        self.api.register(self._authorized, 'authorized', default=True)
        self.api.register(self._authorized_sid, 'authorized_sid', default=True)
        self.api.register(self._preauthorize_sid, 'preauthorize_sid', default=True)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Socks5.namespace)

    def plugin_end(self):
        self.xmpp.remove_handler('Socks5 Bytestreams')
        self.xmpp.remove_handler('Socks5 Streamhost Used')
        self.xmpp['xep_0030'].del_feature(feature=Socks5.namespace)

    def get_socket(self, sid):
        """Returns the socket associated to the SID."""
        return self._sessions.get(sid, None)

    async def handshake(self, to, ifrom=None, sid=None, timeout=None):
        """ Starts the handshake to establish the socks5 bytestreams
        connection.
        """
        if not self._proxies:
            self._proxies = await self.discover_proxies()

        if sid is None:
            sid = uuid4().hex

        used = await self.request_stream(to, sid=sid, ifrom=ifrom, timeout=timeout)
        proxy = used['socks']['streamhost_used']['jid']

        if proxy not in self._proxies:
            log.warning('Received unknown SOCKS5 proxy: %s', proxy)
            return

        try:
            self._sessions[sid] = (await self._connect_proxy(
                    self._get_dest_sha1(sid, self.xmpp.boundjid, to),
                    self._proxies[proxy][0],
                    self._proxies[proxy][1]))[1]
        except socket.error:
            return None
        addr, port = await self._sessions[sid].connected

        # Request that the proxy activate the session with the target.
        await self.activate(proxy, sid, to, timeout=timeout)
        sock = self.get_socket(sid)
        self.xmpp.event('stream:%s:%s' % (sid, to), sock)
        return sock

    def request_stream(self, to, sid=None, ifrom=None, timeout=None, callback=None):
        if sid is None:
            sid = uuid4().hex

        # Requester initiates S5B negotiation with Target by sending
        # IQ-set that includes the JabberID and network address of
        # StreamHost as well as the StreamID (SID) of the proposed
        # bytestream.
        iq = self.xmpp.Iq()
        iq['to'] = to
        iq['from'] = ifrom
        iq['type'] = 'set'
        iq['socks']['sid'] = sid
        for proxy, (host, port) in self._proxies.items():
            iq['socks'].add_streamhost(proxy, host, port)
        return iq.send(timeout=timeout, callback=callback)

    async def discover_proxies(self, jid=None, ifrom=None, timeout=None):
        """Auto-discover the JIDs of SOCKS5 proxies on an XMPP server."""
        if jid is None:
            if self.xmpp.is_component:
                jid = self.xmpp.server
            else:
                jid = self.xmpp.boundjid.server

        discovered = set()

        disco_items = await self.xmpp['xep_0030'].get_items(jid, timeout=timeout)
        disco_items = {item[0] for item in disco_items['disco_items']['items']}

        disco_info_futures = {}
        for item in disco_items:
            disco_info_futures[item] = self.xmpp['xep_0030'].get_info(item, timeout=timeout)

        for item in disco_items:
            try:
                disco_info = await disco_info_futures[item]
            except XMPPError:
                continue
            else:
                # Verify that the identity is a bytestream proxy.
                identities = disco_info['disco_info']['identities']
                for identity in identities:
                    if identity[0] == 'proxy' and identity[1] == 'bytestreams':
                        discovered.add(disco_info['from'])

        for jid in discovered:
            try:
                addr = await self.get_network_address(jid, ifrom=ifrom, timeout=timeout)
                self._proxies[jid] = (addr['socks']['streamhost']['host'],
                                      addr['socks']['streamhost']['port'])
            except XMPPError:
                continue

        return self._proxies

    def get_network_address(self, proxy, ifrom=None, timeout=None, callback=None):
        """Get the network information of a proxy."""
        iq = self.xmpp.Iq(sto=proxy, stype='get', sfrom=ifrom)
        iq.enable('socks')
        return iq.send(timeout=timeout, callback=callback)

    def _get_dest_sha1(self, sid, requester, target):
        # The hostname MUST be SHA1(SID + Requester JID + Target JID)
        # where the output is hexadecimal-encoded (not binary).
        digest = sha1()
        digest.update(sid.encode('utf8'))
        digest.update(str(requester).encode('utf8'))
        digest.update(str(target).encode('utf8'))
        return digest.hexdigest()

    def _handle_streamhost(self, iq):
        """Handle incoming SOCKS5 session request."""
        sid = iq['socks']['sid']
        if not sid:
            raise XMPPError(etype='modify', condition='bad-request')

        if not self._accept_stream(iq):
            raise XMPPError(etype='modify', condition='not-acceptable')

        streamhosts = iq['socks']['streamhosts']
        requester = iq['from']
        target = iq['to']

        dest = self._get_dest_sha1(sid, requester, target)

        proxy_futures = []
        for streamhost in streamhosts:
            proxy_futures.append(self._connect_proxy(
                    dest,
                    streamhost['host'],
                    streamhost['port']))

        async def gather(futures, iq, streamhosts):
            proxies = await asyncio.gather(*futures, return_exceptions=True)
            for streamhost, proxy in zip(streamhosts, proxies):
                if isinstance(proxy, ValueError):
                    continue
                elif isinstance(proxy, socket.error):
                    log.error('Socket error while connecting to the proxy.')
                    continue
                proxy = proxy[1]
                # TODO: what if the future never happens?
                try:
                    addr, port = await proxy.connected
                except socket.error:
                    log.exception('Socket error while connecting to the proxy.')
                    continue
                # TODO: make a better choice than just the first working one.
                used_streamhost = streamhost['jid']
                conn = proxy
                break
            else:
                raise XMPPError(etype='cancel', condition='item-not-found')

            # TODO: close properly the connection to the other proxies.

            iq = iq.reply()
            self._sessions[sid] = conn
            iq['socks']['sid'] = sid
            iq['socks']['streamhost_used']['jid'] = used_streamhost
            iq.send()
            self.xmpp.event('socks5_stream', conn)
            self.xmpp.event('stream:%s:%s' % (sid, requester), conn)

        asyncio.ensure_future(gather(proxy_futures, iq, streamhosts))

    def activate(self, proxy, sid, target, ifrom=None, timeout=None, callback=None):
        """Activate the socks5 session that has been negotiated."""
        iq = self.xmpp.Iq(sto=proxy, stype='set', sfrom=ifrom)
        iq['socks']['sid'] = sid
        iq['socks']['activate'] = target
        return iq.send(timeout=timeout, callback=callback)

    def deactivate(self, sid):
        """Closes the proxy socket associated with this SID."""
        sock = self._sessions.get(sid)
        if sock:
            try:
                # sock.close() will also delete sid from self._sessions (see _connect_proxy)
                sock.close()
            except socket.error:
                pass
            # Though this should not be necessary remove the closed session anyway
            if sid in self._sessions:
                log.warn(('SOCKS5 session with sid = "%s" was not ' +
                          'removed from _sessions by sock.close()') % sid)
                del self._sessions[sid]

    def close(self):
        """Closes all proxy sockets."""
        for sid, sock in self._sessions.items():
            sock.close()
        self._sessions = {}

    def _connect_proxy(self, dest, proxy, proxy_port):
        """ Returns a future to a connection between the client and the server-side
        Socks5 proxy.

        dest : The SHA-1 of (SID + Requester JID + Target JID), in hex. <str>
        host : The hostname or the IP of the proxy. <str>
        port : The port of the proxy. <str> or <int>
        """
        factory = lambda: Socks5Protocol(dest, 0, self.xmpp.event)
        return self.xmpp.loop.create_connection(factory, proxy, proxy_port)

    def _accept_stream(self, iq):
        receiver = iq['to']
        sender = iq['from']
        sid = iq['socks']['sid']

        if self.api['authorized_sid'](receiver, sid, sender, iq):
            return True
        return self.api['authorized'](receiver, sid, sender, iq)

    def _authorized(self, jid, sid, ifrom, iq):
        return self.auto_accept

    def _authorized_sid(self, jid, sid, ifrom, iq):
        log.debug('>>> authed sids: %s', self._preauthed_sids)
        log.debug('>>> lookup: %s %s %s', jid, sid, ifrom)
        if (jid, sid, ifrom) in self._preauthed_sids:
            del self._preauthed_sids[(jid, sid, ifrom)]
            return True
        return False

    def _preauthorize_sid(self, jid, sid, ifrom, data):
        log.debug('>>>> %s %s %s %s', jid, sid, ifrom, data)
        self._preauthed_sids[(jid, sid, ifrom)] = True



# -*- coding: utf-8 -*-
"""
    slixmpp.util.sasl.client
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module was originally based on Dave Cridland's Suelta library.

    Part of Slixmpp: The Slick XMPP Library

    :copryight: (c) 2004-2013 David Alan Cridland
    :copyright: (c) 2013 Nathanael C. Fritz, Lance J.T. Stout

    :license: MIT, see LICENSE for more details
"""

import logging
import stringprep

from slixmpp.util import hashes, bytes, stringprep_profiles


log = logging.getLogger(__name__)


#: Global registry mapping mechanism names to implementation classes.
MECHANISMS = {}


#: Global registry mapping mechanism names to security scores.
MECH_SEC_SCORES = {}


#: The SASLprep profile of stringprep used to validate simple username
#: and password credentials.
saslprep = stringprep_profiles.create(
    nfkc=True,
    bidi=True,
    mappings=[
        stringprep_profiles.b1_mapping,
        stringprep_profiles.c12_mapping],
    prohibited=[
        stringprep.in_table_c12,
        stringprep.in_table_c21,
        stringprep.in_table_c22,
        stringprep.in_table_c3,
        stringprep.in_table_c4,
        stringprep.in_table_c5,
        stringprep.in_table_c6,
        stringprep.in_table_c7,
        stringprep.in_table_c8,
        stringprep.in_table_c9],
    unassigned=[stringprep.in_table_a1])


def sasl_mech(score):
    sec_score = score
    def register(mech):
        n = 0
        mech.score = sec_score
        if mech.use_hashes:
            for hashing_alg in hashes():
                n += 1
                score = mech.score + n
                name = '%s-%s' % (mech.name, hashing_alg)
                MECHANISMS[name] = mech
                MECH_SEC_SCORES[name] = score

                if mech.channel_binding:
                    name += '-PLUS'
                    score += 10
                    MECHANISMS[name] = mech
                    MECH_SEC_SCORES[name] = score
        else:
            MECHANISMS[mech.name] = mech
            MECH_SEC_SCORES[mech.name] = mech.score
            if mech.channel_binding:
                MECHANISMS[mech.name + '-PLUS'] = mech
                MECH_SEC_SCORES[mech.name] = mech.score + 10
        return mech
    return register


class SASLNoAppropriateMechanism(Exception):
    def __init__(self, value=''):
        self.message = value


class SASLCancelled(Exception):
    def __init__(self, value=''):
        self.message = value


class SASLFailed(Exception):
    def __init__(self, value=''):
        self.message = value


class SASLMutualAuthFailed(SASLFailed):
    def __init__(self, value=''):
        self.message = value


class Mech(object):

    name = 'GENERIC'
    score = -1
    use_hashes = False
    channel_binding = False
    required_credentials = set()
    optional_credentials = set()
    security = set()

    def __init__(self, name, credentials, security_settings):
        self.credentials = credentials
        self.security_settings = security_settings
        self.values = {}
        self.base_name = self.name
        self.name = name
        self.setup(name)

    def setup(self, name):
        pass

    def process(self, challenge=b''):
        return b''


def choose(mech_list, credentials, security_settings, limit=None, min_mech=None):
    available_mechs = set(MECHANISMS.keys())
    if limit is None:
        limit = set(mech_list)
    if not isinstance(limit, set):
        limit = set(limit)
    if not isinstance(mech_list, set):
        mech_list = set(mech_list)

    mech_list = mech_list.intersection(limit)
    available_mechs = available_mechs.intersection(mech_list)

    best_score = MECH_SEC_SCORES.get(min_mech, -1)
    best_mech = None
    for name in available_mechs:
        if name in MECH_SEC_SCORES:
            if MECH_SEC_SCORES[name] > best_score:
                best_score = MECH_SEC_SCORES[name]
                best_mech = name
    if best_mech is None:
        raise SASLNoAppropriateMechanism()

    mech_class = MECHANISMS[best_mech]

    try:
        creds = credentials(mech_class.required_credentials,
                            mech_class.optional_credentials)
        for req in mech_class.required_credentials:
            if req not in creds:
                raise SASLCancelled('Missing credential: %s' % req)
        for opt in mech_class.optional_credentials:
            if opt not in creds:
                creds[opt] = b''
        for cred in creds:
            if cred in ('username', 'password', 'authzid'):
                creds[cred] = bytes(saslprep(creds[cred]))
            else:
                creds[cred] = bytes(creds[cred])
        security_opts = security_settings(mech_class.security)

        return mech_class(best_mech, creds, security_opts)
    except SASLCancelled as e:
        log.info('SASL: %s: %s', best_mech, e.message)
        mech_list.remove(best_mech)
        return choose(mech_list, credentials, security_settings,
                limit=limit,
                min_mech=min_mech)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010 Nathanael C. Fritz
    Copyright (C) 2020 "Maxime “pep” Buquet <pep@bouah.net>"
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""
from __future__ import with_statement

import logging
from typing import (
    List,
    Tuple,
    Optional,
)

from slixmpp import (
    Presence,
    Message,
    Iq,
    JID,
)
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin, ET
from slixmpp.xmlstream.handler.callback import Callback
from slixmpp.xmlstream.matcher.xpath import MatchXPath
from slixmpp.xmlstream.matcher.stanzapath import StanzaPath
from slixmpp.xmlstream.matcher.xmlmask import MatchXMLMask
from slixmpp.exceptions import IqError, IqTimeout

from slixmpp.plugins.xep_0045 import stanza
from slixmpp.plugins.xep_0045.stanza import (
    MUCInvite,
    MUCDecline,
    MUCPresence,
    MUCJoin,
    MUCMessage,
    MUCAdminQuery,
    MUCAdminItem,
    MUCHistory,
    MUCOwnerQuery,
    MUCOwnerDestroy,
    MUCStatus,
)


log = logging.getLogger(__name__)

AFFILIATIONS = ('outcast', 'member', 'admin', 'owner', 'none')
ROLES = ('moderator', 'participant', 'visitor', 'none')


class XEP_0045(BasePlugin):

    """
    Implements XEP-0045 Multi-User Chat
    """

    name = 'xep_0045'
    description = 'XEP-0045: Multi-User Chat'
    dependencies = {'xep_0030', 'xep_0004'}
    stanza = stanza

    def plugin_init(self):
        self.rooms = {}
        self.our_nicks = {}
        # load MUC support in presence stanzas
        register_stanza_plugin(MUCMessage, MUCInvite)
        register_stanza_plugin(MUCMessage, MUCDecline)
        register_stanza_plugin(MUCMessage, MUCStatus)
        register_stanza_plugin(MUCPresence, MUCStatus)
        register_stanza_plugin(Presence, MUCPresence)
        register_stanza_plugin(Presence, MUCJoin)
        register_stanza_plugin(MUCJoin, MUCHistory)
        register_stanza_plugin(Message, MUCMessage)
        register_stanza_plugin(Iq, MUCAdminQuery)
        register_stanza_plugin(Iq, MUCOwnerQuery)
        register_stanza_plugin(MUCOwnerQuery, MUCOwnerDestroy)
        register_stanza_plugin(MUCAdminQuery, MUCAdminItem, iterable=True)

        # Register handlers
        self.xmpp.register_handler(
            Callback(
                'MUCPresence',
                MatchXMLMask("<presence xmlns='%s' />" % self.xmpp.default_ns),
                self.handle_groupchat_presence,
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCError',
                MatchXMLMask("<message xmlns='%s' type='error'><error/></message>" % self.xmpp.default_ns),
                self.handle_groupchat_error_message
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCMessage',
                MatchXMLMask("<message xmlns='%s' type='groupchat'><body/></message>" % self.xmpp.default_ns),
                self.handle_groupchat_message
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCSubject',
                MatchXMLMask("<message xmlns='%s' type='groupchat'><subject/></message>" % self.xmpp.default_ns),
                self.handle_groupchat_subject
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCConfig',
                StanzaPath('message/muc/status'),
                self.handle_config_change
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCInvite',
                StanzaPath('message/muc/invite'),
                self.handle_groupchat_invite
        ))
        self.xmpp.register_handler(
            Callback(
                'MUCDecline',
                StanzaPath('message/muc/decline'),
                self.handle_groupchat_decline
        ))

    def plugin_end(self):
        self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)

    def session_bind(self, jid):
        self.xmpp.plugin['xep_0030'].add_feature(stanza.NS)

    def handle_groupchat_invite(self, inv):
        """ Handle an invite into a muc. """
        if inv['from'] not in self.rooms.keys():
            self.xmpp.event("groupchat_invite", inv)

    def handle_groupchat_decline(self, decl):
        """Handle an invitation decline."""
        if decl['from'] in self.room.keys():
            self.xmpp.event('groupchat_decline', decl)

    def handle_config_change(self, msg):
        """Handle a MUC configuration change (with status code)."""
        self.xmpp.event('groupchat_config_status', msg)
        self.xmpp.event('muc::%s::config_status' % msg['from'].bare , msg)

    def handle_groupchat_presence(self, pr):
        """ Handle a presence in a muc.
        """
        got_offline = False
        got_online = False
        if pr['muc']['room'] not in self.rooms.keys():
            return
        self.xmpp.roster[pr['from']].ignore_updates = True
        entry = pr['muc'].get_stanza_values()
        entry['show'] = pr['show'] if pr['show'] in pr.showtypes else None
        entry['status'] = pr['status']
        entry['alt_nick'] = pr['nick']
        if pr['type'] == 'unavailable':
            if entry['nick'] in self.rooms[entry['room']]:
                del self.rooms[entry['room']][entry['nick']]
            got_offline = True
        else:
            if entry['nick'] not in self.rooms[entry['room']]:
                got_online = True
            self.rooms[entry['room']][entry['nick']] = entry
        log.debug("MUC presence from %s/%s : %s", entry['room'],entry['nick'], entry)
        self.xmpp.event("groupchat_presence", pr)
        self.xmpp.event("muc::%s::presence" % entry['room'], pr)
        if got_offline:
            self.xmpp.event("muc::%s::got_offline" % entry['room'], pr)
        if got_online:
            self.xmpp.event("muc::%s::got_online" % entry['room'], pr)

    def handle_groupchat_message(self, msg: Message) -> None:
        """ Handle a message event in a muc.
        """
        self.xmpp.event('groupchat_message', msg)
        self.xmpp.event("muc::%s::message" % msg['from'].bare, msg)

    def handle_groupchat_error_message(self, msg):
        """ Handle a message error event in a muc.
        """
        self.xmpp.event('groupchat_message_error', msg)
        self.xmpp.event("muc::%s::message_error" % msg['from'].bare, msg)


    def handle_groupchat_subject(self, msg: Message) -> None:
        """ Handle a message coming from a muc indicating
        a change of subject (or announcing it when joining the room)
        """
        # See poezio#3452. A message containing subject _and_ (body or thread)
        # is not a subject change.
        if msg['body'] or msg['thread']:
            return None
        self.xmpp.event('groupchat_subject', msg)

    def jid_in_room(self, room: JID, jid: JID) -> bool:
        for nick in self.rooms[room]:
            entry = self.rooms[room][nick]
            if entry is not None and entry['jid'].full == jid:
                return True
        return False

    def get_nick(self, room: JID, jid: JID) -> Optional[str]:
        for nick in self.rooms[room]:
            entry = self.rooms[room][nick]
            if entry is not None and entry['jid'].full == jid:
                return nick

    def join_muc(self, room: JID, nick: str, maxhistory="0", password='',
                 pstatus='', pshow='', pfrom=''):
        """ Join the specified room, requesting 'maxhistory' lines of history.
        """
        stanza = self.xmpp.make_presence(
            pto="%s/%s" % (room, nick), pstatus=pstatus,
            pshow=pshow, pfrom=pfrom
        )
        stanza.enable('muc_join')
        if password:
            stanza['muc_join']['password'] = password
        if maxhistory:
            if maxhistory == "0":
                stanza['muc_join']['history']['maxchars'] = '0'
            else:
                stanza['muc_join']['history']['maxstanzas'] = str(maxhistory)
        self.xmpp.send(stanza)
        self.rooms[room] = {}
        self.our_nicks[room] = nick

    async def destroy(self, room: JID, reason='', altroom='', *,
                      ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """Destroy a room."""
        iq = self.xmpp.make_iq_set(ifrom=ifrom, ito=room)
        iq.enable('mucowner_query')
        iq['mucowner_query'].enable('destroy')
        if altroom:
            iq['mucowner_query']['destroy']['jid'] = altroom
        if reason:
            iq['mucowner_query']['destroy']['reason'] = reason
        await iq.send(**iqkwargs)

    async def set_affiliation(self, room: JID, jid: Optional[JID] = None, nick: Optional[str] = None, *, affiliation: str,
                              ifrom: Optional[JID] = None, **iqkwargs):
        """ Change room affiliation."""
        if affiliation not in AFFILIATIONS:
            raise ValueError('%s is not a valid affiliation' % affiliation)
        if not any((jid, nick)):
            raise ValueError('One of jid or nick must be set')
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        iq.enable('mucadmin_query')
        item = MUCAdminItem()
        item['affiliation'] = affiliation
        if nick:
            item['nick'] = nick
        if jid:
            item['jid'] = jid
        iq['mucadmin_query'].append(item)
        await iq.send(**iqkwargs)

    async def set_role(self, room: JID, nick: str, role: str, *,
                       ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """ Change role property of a nick in a room.
            Typically, roles are temporary (they last only as long as you are in the
            room), whereas affiliations are permanent (they last across groupchat
            sessions).
        """
        if role not in ROLES:
            raise ValueError("Role %s does not exist" % role)
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        iq.enable('mucadmin_query')
        item = MUCAdminItem()
        item['role'] = role
        item['nick'] = nick
        iq['mucadmin_query'].append(item)
        await iq.send(**iqkwargs)

    def invite(self, room: JID, jid: JID, reason: str = '', *,
               mfrom: Optional[JID] = None):
        """ Invite a jid to a room."""
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        msg['muc']['invite']['to'] = jid
        if reason:
            msg['muc']['invite']['reason'] = reason
        self.xmpp.send(msg)

    def decline(self, room: JID, jid: JID, reason: str = '', *,
                mfrom: Optional[JID] = None):
        """Decline a mediated invitation."""
        msg = self.xmpp.make_message(room, mfrom=mfrom)
        msg['muc']['decline']['to'] = jid
        if reason:
            msg['muc']['decline']['reason'] = reason
        self.xmpp.send(msg)

    def leave_muc(self, room: JID, nick: str, msg='', pfrom=None):
        """ Leave the specified room.
        """
        if msg:
            self.xmpp.send_presence(pshow='unavailable', pto="%s/%s" % (room, nick), pstatus=msg, pfrom=pfrom)
        else:
            self.xmpp.send_presence(pshow='unavailable', pto="%s/%s" % (room, nick), pfrom=pfrom)
        del self.rooms[room]

    async def get_room_config(self, room: JID, ifrom=''):
        """Get the room config form in 0004 plugin format """
        iq = self.xmpp.make_iq_get(stanza.NS_OWNER, ito=room, ifrom=ifrom)
        # For now, swallow errors to preserve existing API
        result = await iq.send()
        form = result.xml.find('{http://jabber.org/protocol/muc#owner}query/{jabber:x:data}x')
        if form is None:
            raise ValueError("Configuration form not found")
        return self.xmpp.plugin['xep_0004'].build_form(form)

    async def cancel_config(self, room: JID, *,
                            ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """Cancel a requested config form"""
        query = MUCOwnerQuery()
        x = ET.Element('{jabber:x:data}x', type='cancel')
        query.append(x)
        iq = self.xmpp.make_iq_set(query, ito=room, ifrom=ifrom)
        return await iq.send(**iqkwargs)

    async def set_room_config(self, room: JID, config, *,
                              ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """Send a room config form"""
        query = MUCOwnerQuery()
        config['type'] = 'submit'
        query.append(config)
        iq = self.xmpp.make_iq_set(query, ito=room, ifrom=ifrom)
        return await iq.send(**iqkwargs)

    async def get_affiliation_list(self, room: JID, affiliation: str, *,
                                   ifrom: Optional[JID] = None, **iqkwargs) -> List[JID]:
        """"Get a list of JIDs with the specified affiliation"""
        iq = self.xmpp.make_iq_get(stanza.NS_ADMIN, ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['affiliation'] = affiliation
        result = await iq.send(**iqkwargs)
        return [item['jid'] for item in result['mucadmin_query']]

    async def get_roles_list(self, room: JID, role: str, *,
                             ifrom: Optional[JID] = None, **iqkwargs) -> List[str]:
        """"Get a list of JIDs with the specified role"""
        iq = self.xmpp.make_iq_get(stanza.NS_ADMIN, ito=room, ifrom=ifrom)
        iq['mucadmin_query']['item']['role'] = role
        result = await iq.send(**iqkwargs)
        return [item['nick'] for item in result['mucadmin_query']]

    async def send_affiliation_list(self, room: JID, affiliations: List[Tuple[JID, str]], *,
                                    ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """Send an affiliation delta list"""
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        for jid, affiliation in affiliations:
            item = MUCAdminItem()
            item['jid'] = jid
            item['affiliation'] = affiliation
            iq['mucadmin_query'].append(item)
        return await iq.send(**iqkwargs)

    async def send_role_list(self, room: JID, roles: List[Tuple[str, str]], *,
                            ifrom: Optional[JID] = None, **iqkwargs) -> Iq:
        """Send a role delta list"""
        iq = self.xmpp.make_iq_set(ito=room, ifrom=ifrom)
        for nick, affiliation in roles:
            item = MUCAdminItem()
            item['nick'] = nick
            item['affiliation'] = affiliation
            iq['mucadmin_query'].append(item)
        return await iq.send(**iqkwargs)

    def get_joined_rooms(self) -> List[JID]:
        return self.rooms.keys()

    def get_our_jid_in_room(self, room_jid: JID) -> str:
        """ Return the jid we're using in a room.
        """
        return "%s/%s" % (room_jid, self.our_nicks[room_jid])

    def get_jid_property(self, room, nick, jid_property):
        """ Get the property of a nick in a room, such as its 'jid' or 'affiliation'
            If not found, return None.
        """
        if room in self.rooms and nick in self.rooms[room] and jid_property in self.rooms[room][nick]:
            return self.rooms[room][nick][jid_property]
        else:
            return None

    def get_roster(self, room: JID) -> List[str]:
        """ Get the list of nicks in a room.
        """
        if room not in self.rooms.keys():
            return None
        return self.rooms[room].keys()

    def get_users_by_affiliation(self, room: JID, affiliation='member', *, ifrom: Optional[JID] = None):
        # Preserve old API
        if affiliation not in AFFILIATIONS:
            raise TypeError
        return self.get_affiliation_list(room, affiliation, ifrom=ifrom)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010 Nathanael C. Fritz
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import PluginManager, PluginNotFound, BasePlugin
from slixmpp.plugins.base import register_plugin, load_plugin


__all__ = [
    # XEPS
    'xep_0004',  # Data Forms
    'xep_0009',  # Jabber-RPC
    'xep_0012',  # Last Activity
    'xep_0013',  # Flexible Offline Message Retrieval
#   'xep_0016',  # Privacy Lists. Don’t automatically load
    'xep_0020',  # Feature Negotiation
    'xep_0027',  # Current Jabber OpenPGP Usage
    'xep_0030',  # Service Discovery
    'xep_0033',  # Extended Stanza Addresses
    'xep_0045',  # Multi-User Chat (Client)
    'xep_0047',  # In-Band Bytestreams
#   'xep_0048',  # Legacy Bookmarks. Don’t automatically load
    'xep_0049',  # Private XML Storage
    'xep_0050',  # Ad-hoc Commands
    'xep_0054',  # vcard-temp
    'xep_0059',  # Result Set Management
    'xep_0060',  # Pubsub (Client)
    'xep_0065',  # SOCKS5 Bytestreams
    'xep_0066',  # Out of Band Data
    'xep_0070',  # Verifying HTTP Requests via XMPP
    'xep_0071',  # XHTML-IM
    'xep_0077',  # In-Band Registration
#   'xep_0078',  # Non-SASL auth. Don’t automatically load
    'xep_0079',  # Advanced Message Processing
    'xep_0080',  # User Location
    'xep_0082',  # XMPP Date and Time Profiles
    'xep_0084',  # User Avatar
    'xep_0085',  # Chat State Notifications
    'xep_0086',  # Legacy Error Codes
#   'xep_0091',  # Legacy Delayed Delivery. Don’t automatically load
    'xep_0092',  # Software Version
#   'xep_0095',  # Legacy Stream Initiation. Don’t automatically load
#   'xep_0096',  # Legacy SI File Transfer. Don’t automatically load
    'xep_0106',  # JID Escaping
    'xep_0107',  # User Mood
    'xep_0108',  # User Activity
    'xep_0115',  # Entity Capabilities
    'xep_0118',  # User Tune
    'xep_0122',  # Data Forms Validation
    'xep_0128',  # Extended Service Discovery
    'xep_0131',  # Standard Headers and Internet Metadata
    'xep_0133',  # Service Administration
#   'xep_0138',  # Stream Compression. Broken atm
    'xep_0152',  # Reachability Addresses
    'xep_0153',  # vCard-Based Avatars
    'xep_0163',  # Personal Eventing Protocol
    'xep_0172',  # User Nickname
    'xep_0184',  # Message Receipts
    'xep_0186',  # Invisible Command
    'xep_0191',  # Blocking Command
    'xep_0196',  # User Gaming
    'xep_0198',  # Stream Management
    'xep_0199',  # Ping
    'xep_0202',  # Entity Time
    'xep_0203',  # Delayed Delivery
    'xep_0221',  # Data Forms Media Element
    'xep_0222',  # Persistent Storage of Public Data via Pubsub
    'xep_0223',  # Persistent Storage of Private Data via Pubsub
    'xep_0224',  # Attention
    'xep_0231',  # Bits of Binary
    'xep_0235',  # OAuth Over XMPP
#   'xep_0242',  # XMPP Client Compliance 2009. Don’t automatically load
    'xep_0249',  # Direct MUC Invitations
    'xep_0256',  # Last Activity in Presence
    'xep_0257',  # Client Certificate Management for SASL EXTERNAL
    'xep_0258',  # Security Labels in XMPP
#   'xep_0270',  # XMPP Compliance Suites 2010. Don’t automatically load
    'xep_0279',  # Server IP Check
    'xep_0280',  # Message Carbons
    'xep_0297',  # Stanza Forwarding
    'xep_0300',  # Use of Cryptographic Hash Functions in XMPP
#   'xep_0302',  # XMPP Compliance Suites 2012. Don’t automatically load
    'xep_0308',  # Last Message Correction
    'xep_0313',  # Message Archive Management
    'xep_0319',  # Last User Interaction in Presence
#   'xep_0323',  # IoT Systems Sensor Data. Don’t automatically load
#   'xep_0325',  # IoT Systems Control. Don’t automatically load
    'xep_0332',  # HTTP Over XMPP Transport
    'xep_0333',  # Chat Markers
    'xep_0334',  # Message Processing Hints
    'xep_0335',  # JSON Containers
    'xep_0352',  # Client State Indication
    'xep_0353',  # Jingle Message Initiation
    'xep_0359',  # Unique and Stable Stanza IDs
    'xep_0363',  # HTTP File Upload
    'xep_0369',  # MIX-CORE
    'xep_0377',  # Spam reporting
    'xep_0380',  # Explicit Message Encryption
    'xep_0394',  # Message Markup
    'xep_0403',  # MIX-Presence
    'xep_0404',  # MIX-Anon
    'xep_0405',  # MIX-PAM
    'xep_0421',  # Anonymous unique occupant identifiers for MUCs
    'xep_0422',  # Message Fastening
    'xep_0424',  # Message Retraction
    'xep_0425',  # Message Moderation
    'xep_0428',  # Message Fallback
    'xep_0437',  # Room Activity Indicators
    'xep_0439',  # Quick Response
    'xep_0444',  # Message Reactions
]



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0118 import stanza, UserTune


log = logging.getLogger(__name__)


class XEP_0118(BasePlugin):

    """
    XEP-0118: User Tune
    """

    name = 'xep_0118'
    description = 'XEP-0118: User Tune'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=UserTune.namespace)
        self.xmpp['xep_0163'].remove_interest(UserTune.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('user_tune', UserTune)

    def publish_tune(self, artist=None, length=None, rating=None, source=None,
                     title=None, track=None, uri=None, options=None,
                     ifrom=None, callback=None, timeout=None, timeout_callback=None):
        """
        Publish the user's current tune.

        :param artist: The artist or performer of the song.
        :param length: The length of the song in seconds.
        :param rating: The user's rating of the song (from 1 to 10)
        :param source: The album name, website, or other source of the song.
        :param title: The title of the song.
        :param track: The song's track number, or other unique identifier.
        :param uri: A URL to more information about the song.
        :param options: Optional form of publish options.
        """
        tune = UserTune()
        tune['artist'] = artist
        tune['length'] = length
        tune['rating'] = rating
        tune['source'] = source
        tune['title'] = title
        tune['track'] = track
        tune['uri'] = uri
        return self.xmpp['xep_0163'].publish(tune,
                node=UserTune.namespace,
                options=options,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=timeout_callback)

    def stop(self, ifrom=None, callback=None, timeout=None, timeout_callback=None):
        """
        Clear existing user tune information to stop notifications.
        """
        tune = UserTune()
        return self.xmpp['xep_0163'].publish(tune,
                node=UserTune.namespace,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=timeout_callback)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0186 import stanza
from slixmpp.plugins.xep_0186.stanza import Invisible, Visible
from slixmpp.plugins.xep_0186.invisible_command import XEP_0186


register_plugin(XEP_0186)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""
from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0437.rai import XEP_0437

register_plugin(XEP_0437)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0047 import stanza
from slixmpp.plugins.xep_0047.stanza import Open, Close, Data
from slixmpp.plugins.xep_0047.stream import IBBytestream
from slixmpp.plugins.xep_0047.ibb import XEP_0047


register_plugin(XEP_0047)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.xep_0009.binding import py2xml, xml2py, xml2fault, fault2xml
from threading import RLock
import abc
import inspect
import logging
import slixmpp
import sys
import threading
import traceback

log = logging.getLogger(__name__)

def _isstr(obj):
    return isinstance(obj, str)


# Class decorator to declare a metaclass to a class in a way compatible with Python 2 and 3.
# This decorator is copied from 'six' (https://bitbucket.org/gutworth/six):
#
# Copyright (c) 2010-2015 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
def _add_metaclass(metaclass):
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper

def _intercept(method, name, public):
    def _resolver(instance, *args, **kwargs):
        log.debug("Locally calling %s.%s with arguments %s.", instance.FQN(), method.__name__, args)
        try:
            value = method(instance, *args, **kwargs)
            if value == NotImplemented:
                raise InvocationException("Local handler does not implement %s.%s!" % (instance.FQN(), method.__name__))
            return value
        except InvocationException:
            raise
        except Exception as e:
            raise InvocationException("A problem occurred calling %s.%s!" % (instance.FQN(), method.__name__), e)
    _resolver._rpc = public
    _resolver._rpc_name = method.__name__ if name is None else name
    return _resolver

def remote(function_argument, public = True):
    '''
    Decorator for methods which are remotely callable. This decorator
    works in conjunction with classes which extend ABC Endpoint.
    Example:

        @remote
        def remote_method(arg1, arg2)

    Arguments:
        function_argument -- a stand-in for either the actual method
            OR a new name (string) for the method. In that case the
            method is considered mapped:
            Example:

            @remote("new_name")
            def remote_method(arg1, arg2)

        public -- A flag which indicates if this method should be part
            of the known dictionary of remote methods. Defaults to True.
            Example:

            @remote(False)
            def remote_method(arg1, arg2)

    Note: renaming and revising (public vs. private) can be combined.
    Example:

            @remote("new_name", False)
            def remote_method(arg1, arg2)
    '''
    if hasattr(function_argument, '__call__'):
        return _intercept(function_argument, None, public)
    else:
        if not _isstr(function_argument):
            if not isinstance(function_argument, bool):
                raise Exception('Expected an RPC method name or visibility modifier!')
            else:
                def _wrap_revised(function):
                    function = _intercept(function, None, function_argument)
                    return function
                return _wrap_revised
        def _wrap_remapped(function):
            function = _intercept(function, function_argument, public)
            return function
        return _wrap_remapped


class ACL:
    '''
    An Access Control List (ACL) is a list of rules, which are evaluated
    in order until a match is found. The policy of the matching rule
    is then applied.

    Rules are 3-tuples, consisting of a policy enumerated type, a JID
    expression and a RCP resource expression.

    Examples:
    [ (ACL.ALLOW, '*', '*') ] allow everyone everything, no restrictions
    [ (ACL.DENY, '*', '*') ] deny everyone everything, no restrictions
    [ (ACL.ALLOW, 'test@xmpp.org/unit', 'test.*'),
      (ACL.DENY, '*', '*') ] deny everyone everything, except named
        JID, which is allowed access to endpoint 'test' only.

    The use of wildcards is allowed in expressions, as follows:
    '*' everyone, or everything (= all endpoints and methods)
    'test@xmpp.org/*' every JID regardless of JID resource
    '*@xmpp.org/rpc' every JID from domain xmpp.org with JID res 'rpc'
    'frank@*' every 'frank', regardless of domain or JID res
    'system.*' all methods of endpoint 'system'
    '*.reboot' all methods reboot regardless of endpoint
    '''
    ALLOW = True
    DENY = False

    @classmethod
    def check(cls, rules, jid, resource):
        if rules is None:
            return cls.DENY                  # No rules means no access!
        jid = str(jid)     # Check the string representation of the JID.
        if not jid:
            return cls.DENY                  # Can't check an empty JID.
        for rule in rules:
            policy = cls._check(rule, jid, resource)
            if policy is not None:
                return policy
        return cls.DENY   # By default if not rule matches, deny access.

    @classmethod
    def _check(cls, rule, jid, resource):
        if cls._match(jid, rule[1]) and cls._match(resource, rule[2]):
            return rule[0]
        else:
            return None

    @classmethod
    def _next_token(cls, expression, index):
        new_index = expression.xml.find('*', index)
        if new_index == 0:
            return ''
        else:
            if new_index == -1:
                return expression[index : ]
            else:
                return expression[index : new_index]

    @classmethod
    def _match(cls, value, expression):
        #! print "_match [VALUE] %s [EXPR] %s" % (value, expression)
        index = 0
        position = 0
        while index < len(expression):
            token = cls._next_token(expression, index)
            #! print "[TOKEN] '%s'" % token
            size = len(token)
            if size > 0:
                token_index = value.xml.find(token, position)
                if token_index == -1:
                    return False
                else:
                    #! print "[INDEX-OF] %s" % token_index
                    position = token_index + len(token)
                pass
            if size == 0:
                index += 1
            else:
                index += size
        #! print "index %s position %s" % (index, position)
        return True

ANY_ALL = [ (ACL.ALLOW, '*', '*') ]


class RemoteException(Exception):
    '''
    Base exception for RPC. This exception is raised when a problem
    occurs in the network layer.
    '''

    def __init__(self, message="", cause=None):
        '''
        Initializes a new RemoteException.

        Arguments:
            message -- The message accompanying this exception.
            cause -- The underlying cause of this exception.
        '''
        self._message = message
        self._cause = cause
        pass

    def __str__(self):
        return repr(self._message)

    def get_message(self):
        return self._message

    def get_cause(self):
        return self._cause



class InvocationException(RemoteException):
    '''
    Exception raised when a problem occurs during the remote invocation
    of a method.
    '''
    pass



class AuthorizationException(RemoteException):
    '''
    Exception raised when the caller is not authorized to invoke the
    remote method.
    '''
    pass


class TimeoutException(Exception):
    '''
    Exception raised when the synchronous execution of a method takes
    longer than the given threshold because an underlying asynchronous
    reply did not arrive in time.
    '''
    pass


@_add_metaclass(abc.ABCMeta)
class Callback(object):
    '''
    A base class for callback handlers.
    '''

    @abc.abstractproperty
    def set_value(self, value):
        return NotImplemented

    @abc.abstractproperty
    def cancel_with_error(self, exception):
        return NotImplemented


class Future(Callback):
    '''
    Represents the result of an asynchronous computation.
    '''

    def __init__(self):
        '''
        Initializes a new Future.
        '''
        self._value = None
        self._exception = None
        self._event = threading.Event()
        pass

    def set_value(self, value):
        '''
        Sets the value of this Future. Once the value is set, a caller
        blocked on get_value will be able to continue.
        '''
        self._value = value
        self._event.set()

    def get_value(self, timeout=None):
        '''
        Gets the value of this Future. This call will block until
        the result is available, or until an optional timeout expires.
        When this Future is cancelled with an error,

        Arguments:
            timeout -- The maximum waiting time to obtain the value.
        '''
        self._event.wait(timeout)
        if self._exception:
            raise self._exception
        if not self._event.is_set():
            raise TimeoutException
        return self._value

    def is_done(self):
        '''
        Returns true if a value has been returned.
        '''
        return self._event.is_set()

    def cancel_with_error(self, exception):
        '''
        Cancels the Future because of an error. Once cancelled, a
        caller blocked on get_value will be able to continue.
        '''
        self._exception = exception
        self._event.set()


@_add_metaclass(abc.ABCMeta)
class Endpoint(object):
    '''
    The Endpoint class is an abstract base class for all objects
    participating in an RPC-enabled XMPP network.

    A user subclassing this class is required to implement the method:
        FQN(self)
    where FQN stands for Fully Qualified Name, an unambiguous name
    which specifies which object an RPC call refers to. It is the
    first part in a RPC method name '<fqn>.<method>'.
    '''

    def __init__(self, session, target_jid):
        '''
        Initialize a new Endpoint. This constructor should never be
        invoked by a user, instead it will be called by the factories
        which instantiate the RPC-enabled objects, of which only
        the classes are provided by the user.

        Arguments:
            session -- An RPC session instance.
            target_jid -- the identity of the remote XMPP entity.
        '''
        self.session = session
        self.target_jid = target_jid

    @abc.abstractproperty
    def FQN(self):
        return NotImplemented

    def get_methods(self):
        '''
        Returns a dictionary of all RPC method names provided by this
        class. This method returns the actual  method names as found
        in the class definition which have been decorated with:

            @remote
            def some_rpc_method(arg1, arg2)


        Unless:
            (1) the name has been remapped, in which case the new
                name will be returned.

                    @remote("new_name")
                    def some_rpc_method(arg1, arg2)

            (2) the method is set to hidden

                    @remote(False)
                    def some_hidden_method(arg1, arg2)
        '''
        result = dict()
        for function in dir(self):
            test_attr = getattr(self, function, None)
            try:
                if test_attr._rpc:
                    result[test_attr._rpc_name] = test_attr
            except Exception:
                pass
        return result



class Proxy(Endpoint):
    '''
    Implementation of the Proxy pattern which is intended to wrap
    around Endpoints in order to intercept calls, marshall them and
    forward them to the remote object.
    '''

    def __init__(self, endpoint, callback = None):
        '''
        Initializes a new Proxy.

        Arguments:
            endpoint -- The endpoint which is proxified.
        '''
        self._endpoint = endpoint
        self._callback = callback

    def __getattribute__(self, name, *args):
        if name in ('__dict__', '_endpoint', '_callback'):
            return object.__getattribute__(self, name)
        elif name == 'async':
            return lambda callback: Proxy(self._endpoint, callback)
        else:
            attribute = self._endpoint.__getattribute__(name)
            if hasattr(attribute, '__call__'):
                try:
                    if attribute._rpc:
                        def _remote_call(*args, **kwargs):
                            log.debug("Remotely calling '%s.%s' with arguments %s.", self._endpoint.FQN(), attribute._rpc_name, args)
                            return self._endpoint.session._call_remote(self._endpoint.target_jid, "%s.%s" % (self._endpoint.FQN(), attribute._rpc_name), self._callback, *args, **kwargs)
                        return _remote_call
                except:
                    pass   # If the attribute doesn't exist, don't care!
            return attribute

    def get_endpoint(self):
        '''
        Returns the proxified endpoint.
        '''
        return self._endpoint

    def FQN(self):
        return self._endpoint.FQN()


class JabberRPCEntry(object):


    def __init__(self, endpoint_FQN, call):
        self._endpoint_FQN = endpoint_FQN
        self._call = call

    def call_method(self, args):
        return_value = self._call(*args)
        if return_value is None:
            return return_value
        else:
            return self._return(return_value)

    def get_endpoint_FQN(self):
        return self._endpoint_FQN

    def _return(self, *args):
        return args


class RemoteSession(object):
    '''
    A context object for a Jabber-RPC session.
    '''


    def __init__(self, client, session_close_callback):
        '''
        Initializes a new RPC session.

        Arguments:
            client -- The Slixmpp client associated with this session.
            session_close_callback -- A callback called when the
                session is closed.
        '''
        self._client = client
        self._session_close_callback = session_close_callback
        self._event = threading.Event()
        self._entries = {}
        self._callbacks = {}
        self._acls = {}
        self._lock = RLock()

    def _wait(self):
        self._event.wait()

    def _notify(self, event):
        log.debug("RPC Session as %s started.", self._client.boundjid.full)
        self._client.send_presence()
        self._event.set()
        pass

    def _register_call(self, endpoint, method, name=None):
        '''
        Registers a method from an endpoint as remotely callable.
        '''
        if name is None:
            name = method.__name__
        key = "%s.%s" % (endpoint, name)
        log.debug("Registering call handler for %s (%s).", key, method)
        with self._lock:
            if key in self._entries:
                raise KeyError("A handler for %s has already been regisered!" % endpoint)
            self._entries[key] = JabberRPCEntry(endpoint, method)
        return key

    def _register_acl(self, endpoint, acl):
        log.debug("Registering ACL %s for endpoint %s.", repr(acl), endpoint)
        with self._lock:
            self._acls[endpoint] = acl

    def _register_callback(self, pid, callback):
        with self._lock:
            self._callbacks[pid] = callback

    def forget_callback(self, callback):
        with self._lock:
            pid = self._find_key(self._callbacks, callback)
            if pid is not None:
                del self._callback[pid]
            else:
                raise ValueError("Unknown callback!")
        pass

    def _find_key(self, dict, value):
        """return the key of dictionary dic given the value"""
        search = [k for k, v in dict.items() if v == value]
        if len(search) == 0:
            return None
        else:
            return search[0]

    def _unregister_call(self, key):
        #removes the registered call
        with self._lock:
            if self._entries[key]:
                del self._entries[key]
            else:
                raise ValueError()

    def new_proxy(self, target_jid, endpoint_cls):
        '''
        Instantiates a new proxy object, which proxies to a remote
        endpoint. This method uses a class reference without
        constructor arguments to instantiate the proxy.

        Arguments:
            target_jid -- the XMPP entity ID hosting the endpoint.
            endpoint_cls -- The remote (duck) type.
        '''
        try:
            argspec = inspect.getargspec(endpoint_cls.__init__)
            args = [None] * (len(argspec[0]) - 1)
            result = endpoint_cls(*args)
            Endpoint.__init__(result, self, target_jid)
            return Proxy(result)
        except:
            traceback.print_exc(file=sys.stdout)

    def new_handler(self, acl, handler_cls, *args, **kwargs):
        '''
        Instantiates a new handler object, which is called remotely
        by others. The user can control the effect of the call by
        implementing the remote method in the local endpoint class. The
        returned reference can be called locally and will behave as a
        regular instance.

        Arguments:
            acl -- Access control list (see ACL class)
            handler_clss -- The local (duck) type.
            *args -- Constructor arguments for the local type.
            **kwargs -- Constructor keyworded arguments for the local
                type.
        '''
        argspec = inspect.getargspec(handler_cls.__init__)
        base_argspec = inspect.getargspec(Endpoint.__init__)
        if(argspec == base_argspec):
            result = handler_cls(self, self._client.boundjid.full)
        else:
            result = handler_cls(*args, **kwargs)
            Endpoint.__init__(result, self, self._client.boundjid.full)
        method_dict = result.get_methods()
        for method_name, method in method_dict.items():
            #!!! self._client.plugin['xep_0009'].register_call(result.FQN(), method, method_name)
            self._register_call(result.FQN(), method, method_name)
        self._register_acl(result.FQN(), acl)
        return result

#    def is_available(self, targetCls, pto):
#        return self._client.is_available(pto)

    def _call_remote(self, pto, pmethod, callback, *arguments):
        iq = self._client.plugin['xep_0009'].make_iq_method_call(pto, pmethod, py2xml(*arguments))
        pid = iq['id']
        if callback is None:
            future = Future()
            self._register_callback(pid, future)
            iq.send()
            return future.get_value(30)
        else:
            log.debug("[RemoteSession] _call_remote %s", callback)
            self._register_callback(pid, callback)
            iq.send()

    def close(self, wait=False):
        '''
        Closes this session.
        '''
        self._client.disconnect(wait=wait)
        self._session_close_callback()

    def _on_jabber_rpc_method_call(self, iq):
        iq.enable('rpc_query')
        params = iq['rpc_query']['method_call']['params']
        args = xml2py(params)
        pmethod = iq['rpc_query']['method_call']['method_name']
        try:
            with self._lock:
                entry = self._entries[pmethod]
                rules = self._acls[entry.get_endpoint_FQN()]
            if ACL.check(rules, iq['from'], pmethod):
                return_value = entry.call_method(args)
            else:
                raise AuthorizationException("Unauthorized access to %s from %s!" % (pmethod, iq['from']))
            if return_value is None:
                return_value = ()
            response = self._client.plugin['xep_0009'].make_iq_method_response(iq['id'], iq['from'], py2xml(*return_value))
            response.send()
        except InvocationException as ie:
            fault = dict()
            fault['code'] = 500
            fault['string'] = ie.get_message()
            self._client.plugin['xep_0009']._send_fault(iq, fault2xml(fault))
        except AuthorizationException as ae:
            log.error(ae.get_message())
            error = self._client.plugin['xep_0009']._forbidden(iq)
            error.send()
        except Exception as e:
            if isinstance(e, KeyError):
                log.error("No handler available for %s!", pmethod)
                error = self._client.plugin['xep_0009']._item_not_found(iq)
            else:
                traceback.print_exc(file=sys.stderr)
                log.error("An unexpected problem occurred invoking method %s!", pmethod)
                error = self._client.plugin['xep_0009']._undefined_condition(iq)
            #! print "[REMOTE.PY] _handle_remote_procedure_call AN ERROR SHOULD BE SENT NOW %s " % e
            error.send()

    def _on_jabber_rpc_method_response(self, iq):
        iq.enable('rpc_query')
        args = xml2py(iq['rpc_query']['method_response']['params'])
        pid = iq['id']
        with self._lock:
            callback = self._callbacks[pid]
            del self._callbacks[pid]
        if(len(args) > 0):
            callback.set_value(args[0])
        else:
            callback.set_value(None)
        pass

    def _on_jabber_rpc_method_response2(self, iq):
        iq.enable('rpc_query')
        if iq['rpc_query']['method_response']['fault'] is not None:
            self._on_jabber_rpc_method_fault(iq)
        else:
            args = xml2py(iq['rpc_query']['method_response']['params'])
            pid = iq['id']
            with self._lock:
                callback = self._callbacks[pid]
                del self._callbacks[pid]
            if(len(args) > 0):
                callback.set_value(args[0])
            else:
                callback.set_value(None)
            pass

    def _on_jabber_rpc_method_fault(self, iq):
        iq.enable('rpc_query')
        fault = xml2fault(iq['rpc_query']['method_response']['fault'])
        pid = iq['id']
        with self._lock:
            callback = self._callbacks[pid]
            del self._callbacks[pid]
        e = {
             500: InvocationException
        }[fault['code']](fault['string'])
        callback.cancel_with_error(e)

    def _on_jabber_rpc_error(self, iq):
        pid = iq['id']
        pmethod = self._client.plugin['xep_0009']._extract_method(iq['rpc_query'])
        code = iq['error']['code']
        type = iq['error']['type']
        condition = iq['error']['condition']
        #! print("['REMOTE.PY']._BINDING_handle_remote_procedure_error -> ERROR! ERROR! ERROR! Condition is '%s'" % condition)
        with self._lock:
            callback = self._callbacks[pid]
            del self._callbacks[pid]
        e = {
            'item-not-found': RemoteException("No remote handler available for %s at %s!" % (pmethod, iq['from'])),
            'forbidden': AuthorizationException("Forbidden to invoke remote handler for %s at %s!" % (pmethod, iq['from'])),
            'undefined-condition': RemoteException("An unexpected problem occurred trying to invoke %s at %s!" % (pmethod, iq['from'])),
        }[condition]
        if e is None:
            RemoteException("An unexpected exception occurred at %s!" % iq['from'])
        callback.cancel_with_error(e)


class Remote(object):
    '''
    Bootstrap class for Jabber-RPC sessions. New sessions are openend
    with an existing XMPP client, or one is instantiated on demand.
    '''
    _instance = None
    _sessions = dict()
    _lock = threading.RLock()

    @classmethod
    def new_session_with_client(cls, client, callback=None):
        '''
        Opens a new session with a given client.

        Arguments:
            client -- An XMPP client.
            callback -- An optional callback which can be used to track
                the starting state of the session.
        '''
        with Remote._lock:
            if(client.boundjid.bare in cls._sessions):
                raise RemoteException("There already is a session associated with these credentials!")
            else:
                cls._sessions[client.boundjid.bare] = client

        def _session_close_callback():
            with Remote._lock:
                del cls._sessions[client.boundjid.bare]
        result = RemoteSession(client, _session_close_callback)
        client.plugin['xep_0009'].xmpp.add_event_handler('jabber_rpc_method_call', result._on_jabber_rpc_method_call, threaded=True)
        client.plugin['xep_0009'].xmpp.add_event_handler('jabber_rpc_method_response', result._on_jabber_rpc_method_response, threaded=True)
        client.plugin['xep_0009'].xmpp.add_event_handler('jabber_rpc_method_fault', result._on_jabber_rpc_method_fault, threaded=True)
        client.plugin['xep_0009'].xmpp.add_event_handler('jabber_rpc_error', result._on_jabber_rpc_error, threaded=True)
        if callback is None:
            start_event_handler = result._notify
        else:
            start_event_handler = callback
        client.add_event_handler("session_start", start_event_handler)
        if client.connect():
            client.process(threaded=True)
        else:
            raise RemoteException("Could not connect to XMPP server!")
        pass
        if callback is None:
            result._wait()
        return result

    @classmethod
    def new_session(cls, jid, password, callback=None):
        '''
        Opens a new session and instantiates a new XMPP client.

        Arguments:
            jid -- The XMPP JID for logging in.
            password -- The password for logging in.
            callback -- An optional callback which can be used to track
                the starting state of the session.
        '''
        client = slixmpp.ClientXMPP(jid, password)
        #? Register plug-ins.
        client.register_plugin('xep_0004') # Data Forms
        client.register_plugin('xep_0009') # Jabber-RPC
        client.register_plugin('xep_0030') # Service Discovery
        client.register_plugin('xep_0060') # PubSub
        client.register_plugin('xep_0199') # XMPP Ping
        return cls.new_session_with_client(client, callback)




"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
    This file is part of Slixmpp.

    See the file LICENSE for copying permissio
"""

from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)
from slixmpp.plugins.xep_0422.stanza import ApplyTo
from slixmpp.plugins.xep_0359 import OriginID


NS = 'urn:xmpp:message-retract:0'


class Retract(ElementBase):
    namespace = NS
    name = 'retract'
    plugin_attrib = 'retract'


class Retracted(ElementBase):
    namespace = NS
    name = 'retracted'
    plugin_attrib = 'retracted'
    interfaces = {'stamp'}


def register_plugins():
    register_stanza_plugin(ApplyTo, Retract)
    register_stanza_plugin(Message, Retracted)

    register_stanza_plugin(Retracted, OriginID)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp import Presence
from slixmpp.exceptions import XMPPError
from slixmpp.plugins import BasePlugin, register_plugin
from slixmpp.xmlstream import register_stanza_plugin

from slixmpp.plugins.xep_0012 import stanza, LastActivity


log = logging.getLogger(__name__)


class XEP_0256(BasePlugin):

    name = 'xep_0256'
    description = 'XEP-0256: Last Activity in Presence'
    dependencies = {'xep_0012'}
    stanza = stanza
    default_config = {
        'auto_last_activity': False
    }

    def plugin_init(self):
        register_stanza_plugin(Presence, LastActivity)

        self.xmpp.add_filter('out', self._initial_presence_activity)
        self.xmpp.add_event_handler('connected', self._reset_presence_activity)

        self._initial_presence = set()

    def plugin_end(self):
        self.xmpp.del_filter('out', self._initial_presence_activity)
        self.xmpp.del_event_handler('connected', self._reset_presence_activity)

    def _reset_presence_activity(self, e):
        self._initial_presence = set()

    def _initial_presence_activity(self, stanza):
        if isinstance(stanza, Presence):
            use_last_activity = False

            if self.auto_last_activity and  stanza['show'] in ('xa', 'away'):
                use_last_activity = True

            if stanza['from'] not in self._initial_presence:
                self._initial_presence.add(stanza['from'])
                use_last_activity = True

            if use_last_activity:
                plugin = self.xmpp['xep_0012']
                try:
                    result = plugin.api['get_last_activity'](stanza['from'],
                                                             None,
                                                             stanza['to'])
                    seconds = result['last_activity']['seconds']
                except XMPPError:
                    seconds = None

                if seconds is not None:
                    stanza['last_activity']['seconds'] = seconds
        return stanza


register_plugin(XEP_0256)



# -*- coding: utf-8 -*-
"""
    slixmpp.clientxmpp
    ~~~~~~~~~~~~~~~~~~~~

    This module provides XMPP functionality that
    is specific to client connections.

    Part of Slixmpp: The Slick XMPP Library

    :copyright: (c) 2011 Nathanael C. Fritz
    :license: MIT, see LICENSE for more details
"""

import asyncio
import logging

from slixmpp.jid import JID
from slixmpp.stanza import StreamFeatures
from slixmpp.basexmpp import BaseXMPP
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import XMLStream
from slixmpp.xmlstream.matcher import StanzaPath, MatchXPath
from slixmpp.xmlstream.handler import Callback, CoroutineCallback

# Flag indicating if DNS SRV records are available for use.
try:
    import dns.resolver
except ImportError:
    DNSPYTHON = False
else:
    DNSPYTHON = True


log = logging.getLogger(__name__)


class ClientXMPP(BaseXMPP):

    """
    Slixmpp's client class. (Use only for good, not for evil.)

    Typical use pattern:

    .. code-block:: python

        xmpp = ClientXMPP('user@server.tld/resource', 'password')
        # ... Register plugins and event handlers ...
        xmpp.connect()
        xmpp.process(block=False) # block=True will block the current
                                  # thread. By default, block=False

    :param jid: The JID of the XMPP user account.
    :param password: The password for the XMPP user account.
    :param plugin_config: A dictionary of plugin configurations.
    :param plugin_whitelist: A list of approved plugins that
                    will be loaded when calling
                    :meth:`~slixmpp.basexmpp.BaseXMPP.register_plugins()`.
    :param escape_quotes: **Deprecated.**
    """

    def __init__(self, jid, password, plugin_config=None,
                 plugin_whitelist=None, escape_quotes=True, sasl_mech=None,
                 lang='en', **kwargs):
        if not plugin_whitelist:
            plugin_whitelist = []
        if not plugin_config:
            plugin_config = {}

        BaseXMPP.__init__(self, jid, 'jabber:client', **kwargs)

        self.escape_quotes = escape_quotes
        self.plugin_config = plugin_config
        self.plugin_whitelist = plugin_whitelist
        self.default_port = 5222
        self.default_lang = lang

        self.credentials = {}

        self.password = password

        self.stream_header = "<stream:stream to='%s' %s %s %s %s>" % (
                self.boundjid.host,
                "xmlns:stream='%s'" % self.stream_ns,
                "xmlns='%s'" % self.default_ns,
                "xml:lang='%s'" % self.default_lang,
                "version='1.0'")
        self.stream_footer = "</stream:stream>"

        self.features = set()
        self._stream_feature_handlers = {}
        self._stream_feature_order = []

        self.dns_service = 'xmpp-client'

        #TODO: Use stream state here
        self.authenticated = False
        self.sessionstarted = False
        self.bound = False
        self.bindfail = False

        self.add_event_handler('connected', self._reset_connection_state)
        self.add_event_handler('session_bind', self._handle_session_bind)
        self.add_event_handler('roster_update', self._handle_roster)

        self.register_stanza(StreamFeatures)

        self.register_handler(
                CoroutineCallback('Stream Features',
                     MatchXPath('{%s}features' % self.stream_ns),
                     self._handle_stream_features))
        def roster_push_filter(iq):
            from_ = iq['from']
            if from_ and from_ != JID('') and from_ != self.boundjid.bare:
                reply = iq.reply()
                reply['type'] = 'error'
                reply['error']['type'] = 'cancel'
                reply['error']['code'] = 503
                reply['error']['condition'] = 'service-unavailable'
                reply.send()
                return
            self.event('roster_update', iq)
        self.register_handler(
                Callback('Roster Update',
                         StanzaPath('iq@type=set/roster'),
                         roster_push_filter))

        # Setup default stream features
        self.register_plugin('feature_starttls')
        self.register_plugin('feature_bind')
        self.register_plugin('feature_session')
        self.register_plugin('feature_rosterver')
        self.register_plugin('feature_preapproval')
        self.register_plugin('feature_mechanisms')

        if sasl_mech:
            self['feature_mechanisms'].use_mech = sasl_mech

    @property
    def password(self):
        return self.credentials.get('password', '')

    @password.setter
    def password(self, value):
        self.credentials['password'] = value

    def connect(self, address=tuple(), use_ssl=False,
                force_starttls=True, disable_starttls=False):
        """Connect to the XMPP server.

        When no address is given, a SRV lookup for the server will
        be attempted. If that fails, the server user in the JID
        will be used.

        :param address: A tuple containing the server's host and port.
        :param force_starttls: Indicates that negotiation should be aborted
                               if the server does not advertise support for
                               STARTTLS. Defaults to ``True``.
        :param disable_starttls: Disables TLS for the connection.
                                 Defaults to ``False``.
        :param use_ssl: Indicates if the older SSL connection method
                        should be used. Defaults to ``False``.
        """

        # If an address was provided, disable using DNS SRV lookup;
        # otherwise, use the domain from the client JID with the standard
        # XMPP client port and allow SRV lookup.
        if address:
            self.dns_service = None
        else:
            address = (self.boundjid.host, 5222)
            self.dns_service = 'xmpp-client'

        return XMLStream.connect(self, address[0], address[1], use_ssl=use_ssl,
                                 force_starttls=force_starttls, disable_starttls=disable_starttls)

    def register_feature(self, name, handler, restart=False, order=5000):
        """Register a stream feature handler.

        :param name: The name of the stream feature.
        :param handler: The function to execute if the feature is received.
        :param restart: Indicates if feature processing should halt with
                        this feature. Defaults to ``False``.
        :param order: The relative ordering in which the feature should
                      be negotiated. Lower values will be attempted
                      earlier when available.
        """
        self._stream_feature_handlers[name] = (handler, restart)
        self._stream_feature_order.append((order, name))
        self._stream_feature_order.sort()

    def unregister_feature(self, name, order):
        if name in self._stream_feature_handlers:
            del self._stream_feature_handlers[name]
        self._stream_feature_order.remove((order, name))
        self._stream_feature_order.sort()

    def update_roster(self, jid, **kwargs):
        """Add or change a roster item.

        :param jid: The JID of the entry to modify.
        :param name: The user's nickname for this JID.
        :param subscription: The subscription status. May be one of
                             ``'to'``, ``'from'``, ``'both'``, or
                             ``'none'``. If set to ``'remove'``,
                             the entry will be deleted.
        :param groups: The roster groups that contain this item.
        :param timeout: The length of time (in seconds) to wait
                        for a response before continuing if blocking
                        is used. Defaults to
                        :attr:`~slixmpp.xmlstream.xmlstream.XMLStream.response_timeout`.
        :param callback: Optional reference to a stream handler function.
                         Will be executed when the roster is received.
                         Implies ``block=False``.
        """
        current = self.client_roster[jid]

        name = kwargs.get('name', current['name'])
        subscription = kwargs.get('subscription', current['subscription'])
        groups = kwargs.get('groups', current['groups'])

        timeout = kwargs.get('timeout', None)
        callback = kwargs.get('callback', None)

        return self.client_roster.update(jid, name, subscription, groups,
                                         timeout, callback)

    def del_roster_item(self, jid):
        """Remove an item from the roster.

        This is done by setting its subscription status to ``'remove'``.

        :param jid: The JID of the item to remove.
        """
        return self.client_roster.remove(jid)

    def get_roster(self, callback=None, timeout=None, timeout_callback=None):
        """Request the roster from the server.

        :param callback: Reference to a stream handler function. Will
                         be executed when the roster is received.
        """
        iq = self.Iq()
        iq['type'] = 'get'
        iq.enable('roster')
        if 'rosterver' in self.features:
            iq['roster']['ver'] = self.client_roster.version

        if callback is None:
            callback = lambda resp: self.event('roster_update', resp)
        else:
            orig_cb = callback
            def wrapped(resp):
                self.event('roster_update', resp)
                orig_cb(resp)
            callback = wrapped

        return iq.send(callback, timeout, timeout_callback)

    def _reset_connection_state(self, event=None):
        #TODO: Use stream state here
        self.authenticated = False
        self.sessionstarted = False
        self.bound = False
        self.bindfail = False
        self.features = set()

    async def _handle_stream_features(self, features):
        """Process the received stream features.

        :param features: The features stanza.
        """
        for order, name in self._stream_feature_order:
            if name in features['features']:
                handler, restart = self._stream_feature_handlers[name]
                if asyncio.iscoroutinefunction(handler):
                    result = await handler(features)
                else:
                    result = handler(features)
                if result and restart:
                    # Don't continue if the feature requires
                    # restarting the XML stream.
                    return True
        log.debug('Finished processing stream features.')
        self.event('stream_negotiated')

    def _handle_roster(self, iq):
        """Update the roster after receiving a roster stanza.

        :param iq: The roster stanza.
        """
        if iq['type'] == 'set':
            if iq['from'].bare and iq['from'].bare != self.boundjid.bare:
                raise XMPPError(condition='service-unavailable')

        roster = self.client_roster
        if iq['roster']['ver']:
            roster.version = iq['roster']['ver']
        items = iq['roster']['items']

        valid_subscriptions = ('to', 'from', 'both', 'none', 'remove')
        for jid, item in items.items():
            if item['subscription'] in valid_subscriptions:
                roster[jid]['name'] = item['name']
                roster[jid]['groups'] = item['groups']
                roster[jid]['from'] = item['subscription'] in ('from', 'both')
                roster[jid]['to'] = item['subscription'] in ('to', 'both')
                roster[jid]['pending_out'] = (item['ask'] == 'subscribe')

                roster[jid].save(remove=(item['subscription'] == 'remove'))

        if iq['type'] == 'set':
            resp = self.Iq(stype='result',
                           sto=iq['from'],
                           sid=iq['id'])
            resp.enable('roster')
            resp.send()

    def _handle_session_bind(self, jid):
        """Set the client roster to the JID set by the server.

        :param :class:`slixmpp.xmlstream.jid.JID` jid: The bound JID as
            dictated by the server. The same as :attr:`boundjid`.
        """
        self.client_roster = self.roster[jid]



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2011  Nathanael C. Fritz
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

__all__ = [
    'feature_starttls',
    'feature_mechanisms',
    'feature_bind',
    'feature_session',
    'feature_rosterver',
    'feature_preapproval'
]



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0256.last_activity_presence import XEP_0256, LastActivity

register_plugin(XEP_0256)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp import JID, Iq
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0054 import VCardTemp, stanza
from slixmpp import future_wrapper


log = logging.getLogger(__name__)


class XEP_0054(BasePlugin):

    """
    XEP-0054: vcard-temp
    """

    name = 'xep_0054'
    description = 'XEP-0054: vcard-temp'
    dependencies = {'xep_0030', 'xep_0082'}
    stanza = stanza

    def plugin_init(self):
        """
        Start the XEP-0054 plugin.
        """
        register_stanza_plugin(Iq, VCardTemp)


        self.api.register(self._set_vcard, 'set_vcard', default=True)
        self.api.register(self._get_vcard, 'get_vcard', default=True)
        self.api.register(self._del_vcard, 'del_vcard', default=True)

        self._vcard_cache = {}

        self.xmpp.register_handler(
                Callback('VCardTemp',
                    StanzaPath('iq/vcard_temp'),
                    self._handle_get_vcard))

    def plugin_end(self):
        self.xmpp.remove_handler('VCardTemp')
        self.xmpp['xep_0030'].del_feature(feature='vcard-temp')

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('vcard-temp')

    def make_vcard(self):
        return VCardTemp()

    @future_wrapper
    def get_vcard(self, jid=None, ifrom=None, local=None, cached=False,
                  callback=None, timeout=None, timeout_callback=None):
        if local is None:
            if jid is not None and not isinstance(jid, JID):
                jid = JID(jid)
                if self.xmpp.is_component:
                    if jid.domain == self.xmpp.boundjid.domain:
                        local = True
                else:
                    if str(jid) == str(self.xmpp.boundjid):
                        local = True
                jid = jid.full
            elif jid in (None, ''):
                local = True

        if local:
            vcard = self.api['get_vcard'](jid, None, ifrom)
            if not isinstance(vcard, Iq):
                iq = self.xmpp.Iq()
                if vcard is None:
                    vcard = VCardTemp()
                iq.append(vcard)
                return iq
            return vcard

        if cached:
            vcard = self.api['get_vcard'](jid, None, ifrom)
            if vcard is not None:
                if not isinstance(vcard, Iq):
                    iq = self.xmpp.Iq()
                    iq.append(vcard)
                    return iq
                return vcard

        iq = self.xmpp.Iq()
        iq['to'] = jid
        iq['from'] = ifrom
        iq['type'] = 'get'
        iq.enable('vcard_temp')

        return iq.send(callback=callback, timeout=timeout,
                       timeout_callback=timeout_callback)

    @future_wrapper
    def publish_vcard(self, vcard=None, jid=None, ifrom=None,
                      callback=None, timeout=None, timeout_callback=None):
        self.api['set_vcard'](jid, None, ifrom, vcard)
        if self.xmpp.is_component:
            return

        iq = self.xmpp.Iq()
        iq['to'] = jid
        iq['from'] = ifrom
        iq['type'] = 'set'
        iq.append(vcard)
        return iq.send(callback=callback, timeout=timeout,
                       timeout_callback=timeout_callback)

    def _handle_get_vcard(self, iq):
        if iq['type'] == 'result':
            self.api['set_vcard'](jid=iq['from'], args=iq['vcard_temp'])
            return
        elif iq['type'] == 'get' and self.xmpp.is_component:
            vcard = self.api['get_vcard'](iq['from'].bare)
            if isinstance(vcard, Iq):
                vcard.send()
            else:
                iq = iq.reply()
                iq.append(vcard)
                iq.send()
        elif iq['type'] == 'set':
            raise XMPPError('service-unavailable')

    # =================================================================

    def _set_vcard(self, jid, node, ifrom, vcard):
        self._vcard_cache[jid.bare] = vcard

    def _get_vcard(self, jid, node, ifrom, vcard):
        return self._vcard_cache.get(jid.bare, None)

    def _del_vcard(self, jid, node, ifrom, vcard):
        if jid.bare in self._vcard_cache:
            del self._vcard_cache[jid.bare]



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from base64 import b64encode, b64decode

from slixmpp import JID
from slixmpp.util import bytes
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin


class SecurityLabel(ElementBase):
    name = 'securitylabel'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'security_label'

    def add_equivalent(self, label):
        equiv = EquivalentLabel(parent=self)
        equiv.append(label)
        return equiv


class Label(ElementBase):
    name = 'label'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'label'


class DisplayMarking(ElementBase):
    name = 'displaymarking'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'display_marking'
    interfaces = {'fgcolor', 'bgcolor', 'value'}

    def get_fgcolor(self):
        return self._get_attr('fgcolor', 'black')

    def get_bgcolor(self):
        return self._get_attr('fgcolor', 'white')

    def get_value(self):
        return self.xml.text

    def set_value(self, value):
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''


class EquivalentLabel(ElementBase):
    name = 'equivalentlabel'
    namespace = 'urn:xmpp:sec-label:0'
    plugin_attrib = 'equivalent_label'
    plugin_multi_attrib = 'equivalent_labels'


class Catalog(ElementBase):
    name = 'catalog'
    namespace = 'urn:xmpp:sec-label:catalog:2'
    plugin_attrib = 'security_label_catalog'
    interfaces = {'to', 'from', 'name', 'desc', 'id', 'size', 'restrict'}

    def get_to(self):
        return JID(self._get_attr('to'))
        pass

    def set_to(self, value):
        return self._set_attr('to', str(value))

    def get_from(self):
        return JID(self._get_attr('from'))

    def set_from(self, value):
        return self._set_attr('from', str(value))

    def get_restrict(self):
        value = self._get_attr('restrict', '')
        if value and value.lower() in ('true', '1'):
            return True
        return False

    def set_restrict(self, value):
        self._del_attr('restrict')
        if value:
            self._set_attr('restrict', 'true')
        elif value is False:
            self._set_attr('restrict', 'false')


class CatalogItem(ElementBase):
    name = 'catalog'
    namespace = 'urn:xmpp:sec-label:catalog:2'
    plugin_attrib = 'item'
    plugin_multi_attrib = 'items'
    interfaces = {'selector', 'default'}

    def get_default(self):
        value = self._get_attr('default', '')
        if value.lower() in ('true', '1'):
            return True
        return False

    def set_default(self, value):
        self._del_attr('default')
        if value:
            self._set_attr('default', 'true')
        elif value is False:
            self._set_attr('default', 'false')


class ESSLabel(ElementBase):
    name = 'esssecuritylabel'
    namespace = 'urn:xmpp:sec-label:ess:0'
    plugin_attrib = 'ess'
    interfaces = {'value'}

    def get_value(self):
        if self.xml.text:
            return b64decode(bytes(self.xml.text))
        return ''

    def set_value(self, value):
        self.xml.text = ''
        if value:
            self.xml.text = b64encode(bytes(value))

    def del_value(self):
        self.xml.text = ''


register_stanza_plugin(Catalog, CatalogItem, iterable=True)
register_stanza_plugin(CatalogItem, SecurityLabel)
register_stanza_plugin(EquivalentLabel, ESSLabel)
register_stanza_plugin(Label, ESSLabel)
register_stanza_plugin(SecurityLabel, DisplayMarking)
register_stanza_plugin(SecurityLabel, EquivalentLabel, iterable=True)
register_stanza_plugin(SecurityLabel, Label)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
    This file is part of Slixmpp.

    See the file LICENSE for copying permissio
"""

from slixmpp import JID
from slixmpp.stanza import Iq
from slixmpp.xmlstream import (
    ElementBase,
    register_stanza_plugin,
)

from slixmpp.plugins.xep_0369.stanza import (
    Join,
    Leave,
)

NS = 'urn:xmpp:mix:pam:2'


class ClientJoin(ElementBase):
    namespace = NS
    name = 'client-join'
    plugin_attrib = 'client_join'
    interfaces = {'channel'}


class ClientLeave(ElementBase):
    namespace = NS
    name = 'client-leave'
    plugin_attrib = 'client_leave'
    interfaces = {'channel'}


def register_plugins():
    register_stanza_plugin(Iq, ClientJoin)
    register_stanza_plugin(ClientJoin, Join)

    register_stanza_plugin(Iq, ClientLeave)
    register_stanza_plugin(ClientLeave, Leave)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010 Nathanael C. Fritz, Erik Reuterborg Larsson
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

import slixmpp
from slixmpp.plugins.base import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0080 import stanza, Geoloc


log = logging.getLogger(__name__)


class XEP_0080(BasePlugin):

    """
    XEP-0080: User Location
    """

    name = 'xep_0080'
    description = 'XEP-0080: User Location'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0163'].remove_interest(Geoloc.namespace)
        self.xmpp['xep_0030'].del_feature(feature=Geoloc.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('user_location', Geoloc)

    def publish_location(self, **kwargs):
        """
        Publish the user's current location.

        :param accuracy: Horizontal GPS error in meters.
        :param alt: Altitude in meters above or below sea level.
        :param area: A named area such as a campus or neighborhood.
        :param bearing: GPS bearing (direction in which the entity is
                        heading to reach its next waypoint), measured in
                        decimal degrees relative to true north.
        :param building: A specific building on a street or in an area.
        :param country: The nation where the user is located.
        :param countrycode: The ISO 3166 two-letter country code.
        :param datum: GPS datum.
        :param description: A natural-language name for or description of
                            the location.
        :param error: Horizontal GPS error in arc minutes. Obsoleted by
                      the accuracy parameter.
        :param floor: A particular floor in a building.
        :param lat: Latitude in decimal degrees North.
        :param locality: A locality within the administrative region, such
                         as a town or city.
        :param lon: Longitude in decimal degrees East.
        :param postalcode: A code used for postal delivery.
        :param region: An administrative region of the nation, such
                       as a state or province.
        :param room: A particular room in a building.
        :param speed: The speed at which the entity is moving,
                      in meters per second.
        :param street: A thoroughfare within the locality, or a crossing
                       of two thoroughfares.
        :param text: A catch-all element that captures any other
                     information about the location.
        :param timestamp: UTC timestamp specifying the moment when the
                          reading was taken.
        :param uri: A URI or URL pointing to information about
                    the location.
        :param options: Optional form of publish options.
        """
        options = kwargs.get('options', None)
        ifrom = kwargs.get('ifrom', None)
        callback = kwargs.get('callback', None)
        timeout = kwargs.get('timeout', None)
        timeout_callback = kwargs.get('timeout_callback', None)
        for param in ('ifrom', 'block', 'callback', 'timeout', 'options', 'timeout_callback'):
            if param in kwargs:
                del kwargs[param]

        geoloc = Geoloc()
        geoloc.values = kwargs

        return self.xmpp['xep_0163'].publish(geoloc,
                options=options,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=timeout_callback)

    def stop(self, ifrom=None, callback=None, timeout=None, timeout_callback=None):
        """
        Clear existing user location information to stop notifications.
        """
        geoloc = Geoloc()
        return self.xmpp['xep_0163'].publish(geoloc,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=None)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.xmlstream import ElementBase, ET


class UserActivity(ElementBase):

    name = 'activity'
    namespace = 'http://jabber.org/protocol/activity'
    plugin_attrib = 'activity'
    interfaces = {'value', 'text'}
    sub_interfaces = {'text'}
    general = {'doing_chores', 'drinking', 'eating', 'exercising',
               'grooming', 'having_appointment', 'inactive', 'relaxing',
               'talking', 'traveling', 'undefined', 'working'}
    specific = {'at_the_spa', 'brushing_teeth', 'buying_groceries',
                'cleaning', 'coding', 'commuting', 'cooking', 'cycling',
                'dancing', 'day_off', 'doing_maintenance',
                'doing_the_dishes', 'doing_the_laundry', 'driving',
                'fishing', 'gaming', 'gardening', 'getting_a_haircut',
                'going_out', 'hanging_out', 'having_a_beer',
                'having_a_snack', 'having_breakfast', 'having_coffee',
                'having_dinner', 'having_lunch', 'having_tea', 'hiding',
                'hiking', 'in_a_car', 'in_a_meeting', 'in_real_life',
                'jogging', 'on_a_bus', 'on_a_plane', 'on_a_train',
                'on_a_trip', 'on_the_phone', 'on_vacation',
                'on_video_phone', 'other', 'partying', 'playing_sports',
                'praying', 'reading', 'rehearsing', 'running',
                'running_an_errand', 'scheduled_holiday', 'shaving',
                'shopping', 'skiing', 'sleeping', 'smoking',
                'socializing', 'studying', 'sunbathing', 'swimming',
                'taking_a_bath', 'taking_a_shower', 'thinking',
                'walking', 'walking_the_dog', 'watching_a_movie',
                'watching_tv', 'working_out', 'writing'}

    def set_value(self, value):
        self.del_value()
        general = value
        specific = None
        if isinstance(value, tuple) or isinstance(value, list):
            general = value[0]
            specific = value[1]

        if general in self.general:
            gen_xml = ET.Element('{%s}%s' % (self.namespace, general))
            if specific:
                spec_xml = ET.Element('{%s}%s' % (self.namespace, specific))
                if specific in self.specific:
                    gen_xml.append(spec_xml)
                else:
                    raise ValueError('Unknown specific activity')
            self.xml.append(gen_xml)
        else:
            raise ValueError('Unknown general activity')

    def get_value(self):
        general = None
        specific = None
        gen_xml = None
        for child in self.xml:
            if child.tag.startswith('{%s}' % self.namespace):
                elem_name = child.tag.split('}')[-1]
                if elem_name in self.general:
                    general = elem_name
                    gen_xml = child
        if gen_xml is not None:
            for child in gen_xml:
                if child.tag.startswith('{%s}' % self.namespace):
                    elem_name = child.tag.split('}')[-1]
                    if elem_name in self.specific:
                        specific = elem_name
        return (general, specific)

    def del_value(self):
        curr_value = self.get_value()
        if curr_value[0]:
            self._set_sub_text(curr_value[0], '', keep=False)



"""
    Slixmpp: The Slick XMPP Library
    Implementation of HTTP over XMPP transport
    http://xmpp.org/extensions/xep-0332.html
    Copyright (C) 2015 Riptide IO, sangeeth@riptideio.com
    This file is part of slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.xmlstream import ElementBase


class HTTPData(ElementBase):
    """
    The data element.
    """
    name = 'data'
    namespace = 'urn:xmpp:http'
    interfaces = {'data'}
    plugin_attrib = 'data'
    is_extension = True

    def get_data(self, encoding='text'):
        data = self._get_sub_text(encoding, None)
        return str(data) if data is not None else data

    def set_data(self, data, encoding='text'):
        self._set_sub_text(encoding, text=data)




"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0235 import stanza
from slixmpp.plugins.xep_0235.stanza import OAuth
from slixmpp.plugins.xep_0235.oauth import XEP_0235


register_plugin(XEP_0235)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import asyncio
import os
try:
    from unittest import IsolatedAsyncioTestCase
except ImportError:
    # Python < 3.8
    # just to make sure the imports do not break, but
    # not usable.
    from unittest import TestCase as IsolatedAsyncioTestCase
from typing import (
    List,
)

from slixmpp import JID
from slixmpp.clientxmpp import ClientXMPP


class SlixIntegration(IsolatedAsyncioTestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clients = []
        self.addAsyncCleanup(self._destroy)

    def envjid(self, name):
        """Get a JID from an env var"""
        value = os.getenv(name)
        return JID(value)

    def envstr(self, name):
        """get a str from an env var"""
        return os.getenv(name)

    def register_plugins(self, plugins: List[str]):
        """Register plugins on all known clients"""
        for plugin in plugins:
            for client in self.clients:
                client.register_plugin(plugin)

    def add_client(self, jid: JID, password: str):
        """Register a new client"""
        self.clients.append(ClientXMPP(jid, password))

    async def connect_clients(self):
        """Connect all clients"""
        for client in self.clients:
            client.connect()
        wait = [client.wait_until('session_start') for client in self.clients]
        await asyncio.gather(*wait)

    async def _destroy(self):
        """Kill all clients"""
        for client in self.clients:
            client.abort()



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin, BasePlugin

from slixmpp.plugins.google.gmail import Gmail
from slixmpp.plugins.google.auth import GoogleAuth
from slixmpp.plugins.google.settings import GoogleSettings
from slixmpp.plugins.google.nosave import GoogleNoSave


class Google(BasePlugin):

    """
    Google: Custom GTalk Features

    Also see: <https://developers.google.com/talk/jep_extensions/extensions>
    """

    name = 'google'
    description = 'Google: Custom GTalk Features'
    dependencies = set([
        'gmail',
        'google_settings',
        'google_nosave',
        'google_auth'
    ])

    def __getitem__(self, attr):
        if attr in ('settings', 'nosave', 'auth'):
            return self.xmpp['google_%s' % attr]
        elif attr == 'gmail':
            return self.xmpp['gmail']
        else:
            raise KeyError(attr)


register_plugin(Gmail)
register_plugin(GoogleAuth)
register_plugin(GoogleSettings)
register_plugin(GoogleNoSave)
register_plugin(Google)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.stanza import Message, Presence, Iq
from slixmpp.xmlstream import ElementBase


class Forwarded(ElementBase):
    name = 'forwarded'
    namespace = 'urn:xmpp:forward:0'
    plugin_attrib = 'forwarded'
    interfaces = {'stanza'}

    def get_stanza(self):
        for stanza in self:
            if isinstance(stanza, (Message, Presence, Iq)):
                return stanza
        return ''

    def set_stanza(self, value):
        self.del_stanza()
        self.append(value)

    def del_stanza(self):
        found_stanzas = []
        for stanza in self:
            if isinstance(stanza, (Message, Presence, Iq)):
                found_stanzas.append(stanza)
        for stanza in found_stanzas:
            self.iterables.remove(stanza)
            self.xml.remove(stanza.xml)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.xmlstream import ElementBase, ET


class GoogleAuth(ElementBase):
    name = 'auth'
    namespace = 'http://www.google.com/talk/protocol/auth'
    plugin_attrib = 'google'
    interfaces = {'client_uses_full_bind_result', 'service'}

    discovery_attr= '{%s}client-uses-full-bind-result' % namespace
    service_attr= '{%s}service' % namespace

    def setup(self, xml):
        """Don't create XML for the plugin."""
        self.xml = ET.Element('')

    def get_client_uses_full_bind_result(self):
        return self.parent()._get_attr(self.discovery_attr) == 'true'

    def set_client_uses_full_bind_result(self, value):
        if value in (True, 'true'):
            self.parent()._set_attr(self.discovery_attr, 'true')
        else:
            self.parent()._del_attr(self.discovery_attr)

    def del_client_uses_full_bind_result(self):
        self.parent()._del_attr(self.discovery_attr)

    def get_service(self):
        return self.parent()._get_attr(self.service_attr, '')

    def set_service(self, value):
        if value:
            self.parent()._set_attr(self.service_attr, value)
        else:
            self.parent()._del_attr(self.service_attr)

    def del_service(self):
        self.parent()._del_attr(self.service_attr)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet <mathieui@mathieui.net>
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""
from typing import (
    Iterable,
    Optional,
    Tuple,
)

from slixmpp import JID, Message
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0439 import stanza
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback



class XEP_0439(BasePlugin):
    '''XEP-0439: Quick Response'''

    name = 'xep_0439'
    description = 'Quick Response'
    dependencies = set()
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        stanza.register_plugins()
        self.xmpp.register_handler(Callback(
            'Action received',
            StanzaPath('message/action'),
            self._handle_action,
        ))
        self.xmpp.register_handler(Callback(
            'Response received',
            StanzaPath('message/response'),
            self._handle_response,
        ))
        self.xmpp.register_handler(Callback(
            'ActionSelected received',
            StanzaPath('message/action_selected'),
            self._handle_action_selected,
        ))

    def plugin_send(self):
        self.xmpp.remove_handler('Action received')
        self.xmpp.remove_handler('Response received')
        self.xmpp.remove_handler('ActionSelected received')

    def _handle_response(self, msg: Message):
        self.xmpp.event('responses_received', msg)

    def _handle_action(self, msg: Message):
        self.xmpp.event('action_received', msg)

    def _handle_action_selected(self, msg: Message):
        self.xmpp.event('action_selected', msg)

    def ask_for_response(self, mto: JID, body: str,
                         responses: Iterable[Tuple[str, str]],
                         mtype: str = 'chat', lang: Optional[str] = None, *,
                         mfrom: Optional[JID] = None):
        """
        Send a message with a set of responses.

        :param JID mto: The JID of the entity which will receive the message
        :param str body: The message body of the question
        :param Iterable[Tuple[str, str]] responses: A set of tuples containing
            (value, label) for each response
        :param str mtype: The message type
        :param str lang: The lang of the message (if not use, the default
            for this session will be used.
        """
        if lang is None:
            lang = self.xmpp.default_lang
        msg = self.xmpp.make_message(mto=mto, mfrom=mfrom, mtype=mtype)
        msg['body|%s' % lang] = body
        values = set()
        for value, label in responses:
            if value in values:
                raise ValueError("Duplicate values")
            values.add(value)
            elem = stanza.Response()
            elem['lang'] = lang
            elem['value'] = value
            elem['label'] = label
            msg.append(elem)
        msg.send()

    def ask_for_actions(self, mto: JID, body: str,
                        actions: Iterable[Tuple[str, str]],
                        mtype: str = 'chat', lang: Optional[str] = None, *,
                        mfrom: Optional[JID] = None):
        """
        Send a message with a set of actions.

        :param JID mto: The JID of the entity which will receive the message
        :param str body: The message body of the question
        :param Iterable[Tuple[str, str]] actions: A set of tuples containing
            (action, label) for each action
        :param str mtype: The message type
        :param str lang: The lang of the message (if not use, the default
            for this session will be used.
        """
        if lang is None:
            lang = self.xmpp.default_lang
        msg = self.xmpp.make_message(mto=mto, mfrom=mfrom, mtype=mtype)
        msg['body|%s' % lang] = body
        ids = set()
        for id, label in actions:
            if id in ids:
                raise ValueError("Duplicate ids")
            ids.add(id)
            elem = stanza.Action()
            elem['lang'] = lang
            elem['id'] = id
            elem['label'] = label
            msg.append(elem)
        msg.send()



import asyncio
import socket
import logging

from slixmpp.stanza import Iq
from slixmpp.exceptions import XMPPError


log = logging.getLogger(__name__)


class IBBytestream(object):

    def __init__(self, xmpp, sid, block_size, jid, peer, use_messages=False):
        self.xmpp = xmpp
        self.sid = sid
        self.block_size = block_size
        self.use_messages = use_messages

        if jid is None:
            jid = xmpp.boundjid
        self.self_jid = jid
        self.peer_jid = peer

        self.send_seq = -1
        self.recv_seq = -1

        self.stream_started = False
        self.stream_in_closed = False
        self.stream_out_closed = False

        self.recv_queue = asyncio.Queue()

    async def send(self, data, timeout=None):
        if not self.stream_started or self.stream_out_closed:
            raise socket.error
        if len(data) > self.block_size:
            data = data[:self.block_size]
        self.send_seq = (self.send_seq + 1) % 65536
        seq = self.send_seq
        if self.use_messages:
            msg = self.xmpp.Message()
            msg['to'] = self.peer_jid
            msg['from'] = self.self_jid
            msg['id'] = self.xmpp.new_id()
            msg['ibb_data']['sid'] = self.sid
            msg['ibb_data']['seq'] = seq
            msg['ibb_data']['data'] = data
            msg.send()
        else:
            iq = self.xmpp.Iq()
            iq['type'] = 'set'
            iq['to'] = self.peer_jid
            iq['from'] = self.self_jid
            iq['ibb_data']['sid'] = self.sid
            iq['ibb_data']['seq'] = seq
            iq['ibb_data']['data'] = data
            await iq.send(timeout=timeout)
        return len(data)

    async def sendall(self, data, timeout=None):
        sent_len = 0
        while sent_len < len(data):
            sent_len += await self.send(data[sent_len:self.block_size], timeout=timeout)

    async def sendfile(self, file, timeout=None):
        while True:
            data = file.read(self.block_size)
            if not data:
                break
            await self.send(data, timeout=timeout)

    def _recv_data(self, stanza):
        new_seq = stanza['ibb_data']['seq']
        if new_seq != (self.recv_seq + 1) % 65536:
            self.close()
            raise XMPPError('unexpected-request')
        self.recv_seq = new_seq

        data = stanza['ibb_data']['data']
        if len(data) > self.block_size:
            self.close()
            raise XMPPError('not-acceptable')

        self.recv_queue.put_nowait(data)
        self.xmpp.event('ibb_stream_data', self)

        if isinstance(stanza, Iq):
            stanza.reply().send()

    def recv(self, *args, **kwargs):
        return self.read()

    def read(self):
        if not self.stream_started or self.stream_in_closed:
            raise socket.error
        return self.recv_queue.get_nowait()

    def close(self, timeout=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['to'] = self.peer_jid
        iq['from'] = self.self_jid
        iq['ibb_close']['sid'] = self.sid
        self.stream_out_closed = True
        def _close_stream(_):
            self.stream_in_closed = True
        future = iq.send(timeout=timeout, callback=_close_stream)
        self.xmpp.event('ibb_stream_end', self)
        return future

    def _closed(self, iq):
        self.stream_in_closed = True
        self.stream_out_closed = True
        iq.reply().send()
        self.xmpp.event('ibb_stream_end', self)

    def makefile(self, *args, **kwargs):
        return self

    def connect(*args, **kwargs):
        return None

    def shutdown(self, *args, **kwargs):
        return None



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010  Nathanael C. Fritz
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

import asyncio
# Required for python < 3.7 to use the old ssl implementation
# and manage to do starttls as an unintended side effect
asyncio.sslproto._is_sslproto_available = lambda: False

from slixmpp.stanza import Message, Presence, Iq
from slixmpp.jid import JID, InvalidJID
from slixmpp.xmlstream.stanzabase import ET, ElementBase, register_stanza_plugin
from slixmpp.xmlstream.handler import *
from slixmpp.xmlstream import XMLStream
from slixmpp.xmlstream.matcher import *
from slixmpp.xmlstream.asyncio import asyncio, future_wrapper
from slixmpp.basexmpp import BaseXMPP
from slixmpp.clientxmpp import ClientXMPP
from slixmpp.componentxmpp import ComponentXMPP

from slixmpp.version import __version__, __version_info__



from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class Socks5(ElementBase):
    name = 'query'
    namespace = 'http://jabber.org/protocol/bytestreams'
    plugin_attrib = 'socks'
    interfaces = {'sid', 'activate'}
    sub_interfaces = {'activate'}

    def add_streamhost(self, jid, host, port):
        sh = StreamHost(parent=self)
        sh['jid'] = jid
        sh['host'] = host
        sh['port'] = port


class StreamHost(ElementBase):
    name = 'streamhost'
    namespace = 'http://jabber.org/protocol/bytestreams'
    plugin_attrib = 'streamhost'
    plugin_multi_attrib = 'streamhosts'
    interfaces = {'host', 'jid', 'port'}

    def set_jid(self, value):
        return self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class StreamHostUsed(ElementBase):
    name = 'streamhost-used'
    namespace = 'http://jabber.org/protocol/bytestreams'
    plugin_attrib = 'streamhost_used'
    interfaces = {'jid'}

    def set_jid(self, value):
        return self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


register_stanza_plugin(Socks5, StreamHost, iterable=True)
register_stanza_plugin(Socks5, StreamHostUsed)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2010 Nathanael C. Fritz
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import time
import logging

from typing import Optional, Callable

from slixmpp.jid import JID
from slixmpp.stanza import Iq
from slixmpp import asyncio
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0199 import stanza, Ping


log = logging.getLogger(__name__)


class XEP_0199(BasePlugin):

    """
    XEP-0199: XMPP Ping

    Given that XMPP is based on TCP connections, it is possible for the
    underlying connection to be terminated without the application's
    awareness. Ping stanzas provide an alternative to whitespace based
    keepalive methods for detecting lost connections.

    Also see <http://www.xmpp.org/extensions/xep-0199.html>.

    Attributes:
        keepalive -- If True, periodically send ping requests
                     to the server. If a ping is not answered,
                     the connection will be reset.
        interval  -- Time in seconds between keepalive pings.
                     Defaults to 300 seconds.
        timeout   -- Time in seconds to wait for a ping response.
                     Defaults to 30 seconds.
    Methods:
        send_ping -- Send a ping to a given JID, returning the
                     round trip time.
    """

    name = 'xep_0199'
    description = 'XEP-0199: XMPP Ping'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'keepalive': False,
        'interval': 300,
        'timeout': 30
    }

    def plugin_init(self):
        """
        Start the XEP-0199 plugin.
        """

        register_stanza_plugin(Iq, Ping)

        self.xmpp.register_handler(
                Callback('Ping',
                         StanzaPath('iq@type=get/ping'),
                         self._handle_ping))

        if self.keepalive:
            self.xmpp.add_event_handler('session_start',
                                        self.enable_keepalive)
            self.xmpp.add_event_handler('session_end',
                                        self.disable_keepalive)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Ping.namespace)
        self.xmpp.remove_handler('Ping')
        if self.keepalive:
            self.xmpp.del_event_handler('session_start',
                                        self.enable_keepalive)
            self.xmpp.del_event_handler('session_end',
                                        self.disable_keepalive)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Ping.namespace)

    def enable_keepalive(self, interval=None, timeout=None):
        if interval:
            self.interval = interval
        if timeout:
            self.timeout = timeout

        self.keepalive = True
        handler = lambda event=None: asyncio.ensure_future(
            self._keepalive(event),
            loop=self.xmpp.loop,
        )
        self.xmpp.schedule('Ping keepalive',
                           self.interval,
                           handler,
                           repeat=True)

    def disable_keepalive(self, event=None):
        self.xmpp.cancel_schedule('Ping keepalive')

    async def _keepalive(self, event=None):
        log.debug("Keepalive ping...")
        try:
            rtt = await self.ping(self.xmpp.boundjid.host, timeout=self.timeout)
        except IqTimeout:
            log.debug("Did not receive ping back in time. " + \
                      "Requesting Reconnect.")
            self.xmpp.reconnect(0.0, "Ping timeout after %ds" % self.timeout)
        else:
            log.debug('Keepalive RTT: %s' % rtt)

    def _handle_ping(self, iq):
        """Automatically reply to ping requests."""
        log.debug("Pinged by %s", iq['from'])
        iq.reply().send()

    def send_ping(self, jid: JID, ifrom: Optional[JID] = None,
                  timeout: Optional[int] = None,
                  callback: Optional[Callable] = None,
                  timeout_callback: Optional[Callable] = None):
        """Send a ping request.

        :param jid: The JID that will receive the ping.
        """
        if not timeout:
            timeout = self.timeout

        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['to'] = jid
        iq['from'] = ifrom
        iq.enable('ping')

        return iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)

    async def ping(self, jid: Optional[JID] =None,
            ifrom: Optional[JID] = None, timeout: Optional[int] = None) -> float:
        """Send a ping request and calculate RTT.
        This is a coroutine.

        :param jid: The JID that will receive the ping.
        """
        own_host = False
        if not jid:
            if self.xmpp.is_component:
                jid = self.xmpp.server
            else:
                jid = self.xmpp.boundjid.host
        jid = JID(jid)
        if jid == self.xmpp.boundjid.host or \
                self.xmpp.is_component and jid == self.xmpp.server:
            own_host = True

        if not timeout:
            timeout = self.timeout

        start = time.time()

        log.debug('Pinging %s' % jid)
        try:
            await self.send_ping(jid, ifrom=ifrom, timeout=timeout)
        except IqError as e:
            if own_host:
                rtt = time.time() - start
                log.debug('Pinged %s, RTT: %s', jid, rtt)
                return rtt
            else:
                raise e
        else:
            rtt = time.time() - start
            log.debug('Pinged %s, RTT: %s', jid, rtt)
            return rtt



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0377.stanza import Report, Text
from slixmpp.plugins.xep_0377.spam_reporting import XEP_0377


register_plugin(XEP_0377)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp import JID
from typing import Dict, List, Optional, Callable
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0152 import stanza, Reachability
from slixmpp.plugins.xep_0004 import Form


log = logging.getLogger(__name__)


class XEP_0152(BasePlugin):

    """
    XEP-0152: Reachability Addresses
    """

    name = 'xep_0152'
    description = 'XEP-0152: Reachability Addresses'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Reachability.namespace)
        self.xmpp['xep_0163'].remove_interest(Reachability.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('reachability', Reachability)

    def publish_reachability(self, addresses: List[Dict[str, str]],
                             options: Optional[Form] = None,
                             ifrom: Optional[JID] = None,
                             callback: Optional[Callable] = None,
                             timeout: Optional[int] = None,
                             timeout_callback: Optional[Callable] = None):
        """
        Publish alternative addresses where the user can be reached.

        :param addresses: A list of dictionaries containing the URI and
                          optional description for each address.
        :param options: Optional form of publish options.
        """
        if not isinstance(addresses, (list, tuple)):
            addresses = [addresses]
        reach = Reachability()
        for address in addresses:
            if not hasattr(address, 'items'):
                address = {'uri': address}

            addr = stanza.Address()
            for key, val in address.items():
                addr[key] = val
            reach.append(addr)
        return self.xmpp['xep_0163'].publish(reach,
                node=Reachability.namespace,
                options=options,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=timeout_callback)

    def stop(self, ifrom: Optional[JID] = None,
             callback: Optional[Callable] = None,
             timeout: Optional[int] = None,
             timeout_callback: Optional[Callable] = None):
        """
        Clear existing user activity information to stop notifications.
        """
        reach = Reachability()
        return self.xmpp['xep_0163'].publish(reach,
                node=Reachability.namespace,
                ifrom=ifrom,
                callback=callback,
                timeout=timeout,
                timeout_callback=timeout_callback)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0049.stanza import PrivateXML
from slixmpp.plugins.xep_0049.private_storage import XEP_0049


register_plugin(XEP_0049)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permissio
"""

from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0313.stanza import Result, MAM, Preferences
from slixmpp.plugins.xep_0313.mam import XEP_0313


register_plugin(XEP_0313)



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import logging

from slixmpp import ClientXMPP
from slixmpp.stanza import StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0352 import stanza, Active, Inactive, ClientStateIndication


log = logging.getLogger(__name__)


class XEP_0352(BasePlugin):

    """
    XEP-0352: Client State Indication
    """

    name = 'xep_0352'
    description = 'XEP-0352: Client State Indication'
    dependencies = set()
    stanza = stanza
    default_config = {
        "order": 12000,
    }

    def plugin_init(self):
        """Start the XEP-0352 plugin."""

        self.enabled = False

        register_stanza_plugin(StreamFeatures, ClientStateIndication)
        self.xmpp.register_stanza(stanza.Active)
        self.xmpp.register_stanza(stanza.Inactive)

        if isinstance(self.xmpp, ClientXMPP):
            self.xmpp.register_feature('csi',
                    self._handle_csi_feature,
                    restart=False,
                    order=self.order)


    def plugin_end(self):
        if self.xmpp.is_component:
            return

        if isinstance(self.xmpp, ClientXMPP):
            self.xmpp.unregister_feature('csi', self.order)
        self.xmpp.remove_stanza(stanza.Active)
        self.xmpp.remove_stanza(stanza.Inactive)

    def send_active(self):
        """Send an 'active' state"""
        if self.enabled:
            self.xmpp.send_raw(str(stanza.Active(self.xmpp)))

    def send_inactive(self):
        """Send an 'active' state"""
        if self.enabled:
            self.xmpp.send_raw(str(stanza.Inactive(self.xmpp)))

    def _handle_csi_feature(self, features):
        """
        Enable CSI
        """
        if 'csi' in self.xmpp.features:
            log.debug('CSI already enabled')
            return False
        self.enabled = True
        self.xmpp.event('csi_enabled', features)
        return False



"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2020 Mathieu Pasquet
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

from typing import Set, Iterable
from slixmpp.xmlstream import ElementBase
try:
    from emoji import UNICODE_EMOJI
except ImportError:
    UNICODE_EMOJI = None


NS = 'urn:xmpp:reactions:0'

class Reactions(ElementBase):
    name = 'reactions'
    plugin_attrib = 'reactions'
    namespace = NS
    interfaces = {'id', 'values'}

    def get_values(self, *, all_chars=False) -> Set[str]:
        """"Get all reactions as str"""
        reactions = set()
        for reaction in self:
            value = reaction['value']
            if UNICODE_EMOJI and not all_chars:
                if value in UNICODE_EMOJI:
                    reactions.add(reaction['value'])
            else:
                reactions.add(reaction['value'])
        return reactions

    def set_values(self, values: Iterable[str], *, all_chars=False):
        """"Set all reactions as str"""
        for element in self.xml.findall('reaction'):
            self.xml.remove(element)
        for reaction_txt in values:
            reaction = Reaction()
            reaction.set_value(reaction_txt, all_chars=all_chars)
            self.append(reaction)


class Reaction(ElementBase):
    name = 'reaction'
    namespace = NS
    interfaces = {'value'}

    def get_value(self) -> str:
        return self.xml.text

    def set_value(self, value: str, *, all_chars=False):
        if UNICODE_EMOJI and not all_chars:
            if not value in UNICODE_EMOJI:
                raise ValueError("%s is not a valid emoji" % value)
        self.xml.text = value




"""
    Slixmpp: The Slick XMPP Library
    Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
    This file is part of Slixmpp.

    See the file LICENSE for copying permission.
"""

import hashlib
import logging

from slixmpp.stanza import Presence
from slixmpp.exceptions import XMPPError, IqTimeout
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0153 import stanza, VCardTempUpdate
from slixmpp import asyncio, future_wrapper


log = logging.getLogger(__name__)


class XEP_0153(BasePlugin):

    name = 'xep_0153'
    description = 'XEP-0153: vCard-Based Avatars'
    dependencies = {'xep_0054'}
    stanza = stanza

    def plugin_init(self):
        self._hashes = {}

        register_stanza_plugin(Presence, VCardTempUpdate)

        self.xmpp.add_filter('out', self._update_presence)

        self.xmpp.add_event_handler('session_start', self._start)
        self.xmpp.add_event_handler('session_end', self._end)

        self.xmpp.add_event_handler('presence_available', self._recv_presence)
        self.xmpp.add_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.add_event_handler('presence_xa', self._recv_presence)
        self.xmpp.add_event_handler('presence_chat', self._recv_presence)
        self.xmpp.add_event_handler('presence_away', self._recv_presence)

        self.api.register(self._set_hash, 'set_hash', default=True)
        self.api.register(self._get_hash, 'get_hash', default=True)
        self.api.register(self._reset_hash, 'reset_hash', default=True)

    def plugin_end(self):
        self.xmpp.del_filter('out', self._update_presence)
        self.xmpp.del_event_handler('session_start', self._start)
        self.xmpp.del_event_handler('session_end', self._end)
        self.xmpp.del_event_handler('presence_available', self._recv_presence)
        self.xmpp.del_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.del_event_handler('presence_xa', self._recv_presence)
        self.xmpp.del_event_handler('presence_chat', self._recv_presence)
        self.xmpp.del_event_handler('presence_away', self._recv_presence)

    @future_wrapper
    def set_avatar(self, jid=None, avatar=None, mtype=None, timeout=None,
                   callback=None, timeout_callback=None):
        if jid is None:
            jid = self.xmpp.boundjid.bare

        future = asyncio.Future()

        def propagate_timeout_exception(fut):
            try:
                fut.done()
            except IqTimeout as e:
                future.set_exception(e)

        def custom_callback(result):
            vcard = result['vcard_temp']
            vcard['PHOTO']['TYPE'] = mtype
            vcard['PHOTO']['BINVAL'] = avatar

            new_future = self.xmpp['xep_0054'].publish_vcard(jid=jid,
                                                             vcard=vcard,
                                                             timeout=timeout,
                                                             callback=next_callback,
                                                             timeout_callback=timeout_callback)
            new_future.add_done_callback(propagate_timeout_exception)

        def next_callback(result):
            if result['type'] == 'error':
                future.set_exception(result)
            else:
                self.api['reset_hash'](jid)
                self.xmpp.roster[jid].send_last_presence()

                future.set_result(result)

        first_future = self.xmpp['xep_0054'].get_vcard(jid, cached=False, timeout=timeout,
                                                       callback=custom_callback,
                                                       timeout_callback=timeout_callback)
        first_future.add_done_callback(propagate_timeout_exception)
        return future

    async def _start(self, event):
        try:
            vcard = await self.xmpp['xep_0054'].get_vcard(self.xmpp.boundjid.bare)
            data = vcard['vcard_temp']['PHOTO']['BINVAL']
            if not data:
                new_hash = ''
            else:
                new_hash = hashlib.sha1(data).hexdigest()
            self.api['set_hash'](self.xmpp.boundjid, args=new_hash)
        except XMPPError:
            log.debug('Could not retrieve vCard for %s', self.xmpp.boundjid.bare)

    def _end(self, event):
        pass

    def _update_presence(self, stanza):
        if not isinstance(stanza, Presence):
            return stanza

        if stanza['type'] not in ('available', 'dnd', 'chat', 'away', 'xa'):
            return stanza

        current_hash = self.api['get_hash'](stanza['from'])
        stanza['vcard_temp_update']['photo'] = current_hash
        return stanza

    def _reset_hash(self, jid, node, ifrom, args):
        own_jid = (jid.bare == self.xmpp.boundjid.bare)
        if self.xmpp.is_component:
            own_jid = (jid.domain == self.xmpp.boundjid.domain)

        self.api['set_hash'](jid, args=None)
        if own_jid:
            self.xmpp.roster[jid].send_last_presence()

        def callback(iq):
            if iq['type'] == 'error':
                log.debug('Could not retrieve vCard for %s', jid)
                return
            try:
                data = iq['vcard_temp']['PHOTO']['BINVAL']
            except ValueError:
                log.debug('Invalid BINVAL in vCard’s PHOTO for %s:', jid, exc_info=True)
                data = None
            if not data:
                new_hash = ''
            else:
                new_hash = hashlib.sha1(data).hexdigest()

            self.api['set_hash'](jid, args=new_hash)

        self.xmpp['xep_0054'].get_vcard(jid=jid.bare, ifrom=ifrom,
                                        callback=callback)

    def _recv_presence(self, pres):
        try:
            if pres['muc']['affiliation']:
                # Don't process vCard avatars for MUC occupants
                # since they all share the same bare JID.
                return
        except: pass

        if not pres.match('presence/vcard_temp_update'):
            self.api['set_hash'](pres['from'], args=None)
            return

        data = pres['vcard_temp_update']['photo']
        if data is None:
            return
        self.xmpp.event('vcard_avatar_update', pres)

    # =================================================================

    def _get_hash(self, jid, node, ifrom, args):
        return self._hashes.get(jid.bare, None)

    def _set_hash(self, jid, node, ifrom, args):
        self._hashes[jid.bare] = args
